# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import time
from agent.lib import filelock
from agent.lib.constants import LOCAL_SERVER_DB, CLUSTER_LOCK, SLEEP_ON_ERR
from agent.lib.config import get_config
from agent.lib.db import read_item, write_item, dump_db, get_cluster_db_home
from agent.lib.notification import notify_manager
from agent.lib.settings import is_master
from agent.target.cluster import is_clustered
from agent.target.dlm import acquire_master_dlm_lock, release_master_dlm_lock
from agent.target.host import (host_alive, ifconfig_vip_up, ifconfig_vip_down,
                             is_holding_vip)


def setup_master_env():
    """Setup master environment:
        1) setup master vip.
        2) store master info to master db.
    """
    d = dump_db(LOCAL_SERVER_DB)
    db_home = get_cluster_db_home()

    vip = read_item("server_pool", "pool_virtual_ip", db_home)
    ip = d.get("registered_ip")
    ifconfig_vip_up(vip, ip)

    my_hostname = d.get("registered_hostname")
    write_item("server_pool", "pool_master_hostname", my_hostname, db_home)


def teardown_master_env():
    """Teardown master environment:
        1) release the master_vip
        2) release master dlm lock
    """
    vip = read_item("server_pool", "pool_virtual_ip", get_cluster_db_home())
    if vip:
        ifconfig_vip_down(vip)

    release_master_dlm_lock()


def send_remaster_event():
    hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
    msg = "{CLUSTER} {REMASTER} master is [%s]" % hostname
    try:
        notify_manager(msg)
    except Exception, e:
        logging.debug("Error sending notification: %s", e)


take_crown = True


def cluster_remaster():
    global take_crown
    db_home = get_cluster_db_home()
    pool_vip = read_item("server_pool", "pool_virtual_ip", db_home)
    if not pool_vip:
        logging.debug("VIP not configured")
        return
    if acquire_master_dlm_lock():
        if host_alive(pool_vip):
            if is_holding_vip(pool_vip):
                if is_master():
                    write_item(LOCAL_SERVER_DB, "is_master", True)
                    if take_crown:
                        take_crown = False
                        logging.info("It is master, VIP=%s", pool_vip)
                else:
                    take_crown = True
                    ifconfig_vip_down(pool_vip)
                    write_item(LOCAL_SERVER_DB, "is_master", False)
                    release_master_dlm_lock()
                    logging.debug("It is not master: VIP released")
            else:
                take_crown = True
                write_item(LOCAL_SERVER_DB, "is_master", False)
                release_master_dlm_lock()
                logging.debug("VIP in use: master lock released")
        else:
            take_crown = True
            try:
                setup_master_env()
                logging.debug("Master server elected: setup done")
                write_item(LOCAL_SERVER_DB, "is_master", True)
                send_remaster_event()
            finally:
                if not is_master():
                    release_master_dlm_lock()
    else:
        take_crown = True
        write_item(LOCAL_SERVER_DB, "is_master", False)


def remaster():
    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        d = dump_db(LOCAL_SERVER_DB)
        if is_clustered():
            cluster_remaster()
        else:
            vip = d.get("pool_virtual_ip")
            ip = d.get("registered_ip")
            master_flag = d.get("is_master")
            if master_flag and vip and not is_holding_vip(vip):
                if host_alive(vip):
                    logging.debug("VIP %s in use: conflict detected. Clearing master flag", vip)
                    try:
                        write_item(LOCAL_SERVER_DB, "is_master", False)
                    except Exception, e1:
                        logging.exception("failed to clear master flag - error %s", e1)
                    else:
                        logging.debug("master flag cleared")
                        try:
                            msg = "{CLUSTER} {WARNING_POOL} Could not obtain VIP [%s]" % vip
                            notify_manager(msg)
                        except Exception, e2:
                            logging.exception("failed to notify manager - error %s", e2)
                else:
                    ifconfig_vip_up(vip, ip)
                    logging.info("It is master, VIP=%s", vip)
                    send_remaster_event()
    finally:
        lock.release()


def teardown_master_server():
    if is_clustered():
        teardown_master_env()
    else:
        vip = read_item(LOCAL_SERVER_DB, "pool_virtual_ip")
        if vip:
            ifconfig_vip_down(vip)


def serve_forever():
    logging.info("REMASTER SERVER STARTED")
    config = get_config()
    remaster_interval = int(config.get("remaster", "interval"))
    while True:
        try:
            time.sleep(remaster_interval)
            remaster()
        except Exception, e:
            logging.exception("Error in remaster process: %s", e)
            time.sleep(SLEEP_ON_ERR)


if __name__ == '__main__':
    serve_forever()
