# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import subprocess
import stat
import time

from agent.extension import OVSDiscoverExtension
from agent.lib.config import get_config


def get_ovs_info():
    ovs_info = {}
    f = open("/etc/ovs-info")
    for line in f:
        m = re.match(r'^release:(.*)$', line)
        if m:
            ovs_info['release'] = m.group(1).strip()
        m = re.match(r'^date:(.*)$', line)
        if m:
            ovs_info['date'] = m.group(1).strip()
        m = re.match(r'^build:(.*)$', line)
        if m:
            ovs_info['build'] = m.group(1).strip()
    f.close()
    return ovs_info


def run_cmd(cmd, success_return_code=(0,)):
    if not isinstance(cmd, list):
        raise Exception("Only accepts list!")
    cmd = [str(x) for x in cmd]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    (stdoutdata, stderrdata) = proc.communicate()
    if proc.returncode not in success_return_code:
        raise RuntimeError('Command: %s failed (%s): stderr: %s stdout: %s'
                           % (cmd, proc.returncode, stderrdata, stdoutdata))
    return str(stdoutdata)


def retry_cmd(cmd, interval=5, count=3):
    # retry if failed
    for i in range(1, count):
        try:
            output = run_cmd(cmd)
            return output
        except:
            time.sleep(interval)
    # retry for the last time
    return run_cmd(cmd)


def get_server_unique_id():
    """
    Obtain a globally unique identifier for this server.

    This invokes legacy C code from the VI agent and returns a 16 byte
    unique ID in the form of a string of hex digits:
        xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx:xx

    @return    The UUID string or None on error
    """
    config = get_config()
    try:
        fakeuuid = config.get("server","fakeuuid")
    except:
        fakeuuid = ""

    if (fakeuuid != ""):
        response = fakeuuid
        return response
    else:
        (status, response, logmsg) = OVSDiscoverExtension.unique_id(None)
        if (status == 0):
            return response
        else:
            logging.error(logmsg)
            return None


def set_nice(increment=10, io_cls=2, io_data=7):
    os.nice(increment)
    run_cmd(['ionice', '-c', io_cls, '-n', io_data, '-p', os.getpid()])


def reset_console_access():
    run_cmd(['service', 'ovm-consoled', 'restart'])


def guess_mimetype(filename):
    ftype = run_cmd(['file', '-b', '-i', filename])
    ftype = ftype.split(',')[0] # application/x-tar, POSIX
    ftype = ftype.split(';')[0] # application/x-tar; charset=binary
    return ftype.strip()


def host_alive(ip):
    status = os.popen("ping -q -c 2 %s" % ip).close()
    return not status


def get_netmask(ip_address):
    """
    Return the netmask of the network where the specified IP address
    is defined. The IP address should be defined on a network interface
    of the server, otherwise None is returned.

    """
    output = run_cmd(["ifconfig"]).splitlines()
    for line in output:
        if (':' + ip_address + ' ') in line:
            netmask = line.split(':')[3]
            return netmask

    return None


def _get_ip_dev(ip):
    """Get dev the ip is bound to."""
    cmd = ["ip", "addr", "show", "to", ip]
    output = run_cmd(cmd).splitlines()
    for line in output:
        if "scope global" in line:
            return line.split()[-1]
    return ''


def ifconfig_vip_up(vip, ip):
    """Bind vip to dev that ip is assigned to."""
    dev = _get_ip_dev(vip)
    if dev: # already bound
        return
    cmd = ["ip", "addr", "show", "to", ip]
    output = run_cmd(cmd)
    lines = output.strip().splitlines()
    if len(lines) < 2:
        raise Exception("IP %s is not up." % ip)
    parts = lines[-1].split()
    dev = parts[-1]
    netmask = parts[1].split("/")[-1]

    dev_ip = dev
    if ":" in dev:
        dev_ip = dev.split(":")[0]

    cmd = ["ifconfig"]
    output = run_cmd(cmd).splitlines()
    dev_list = []
    for line in output:
        if "Link encap" in line:
            dev_list.append(line.split()[0])

    idx = 0
    while True:
        dev_vip = "%s:%s" % (dev_ip, idx)
        if dev_vip not in dev_list:
            break
        else:
            idx += 1

    cmd = ["ifconfig", dev_vip, "%s/%s" % (vip, netmask)]
    run_cmd(cmd)

    #refresh switch arp table
    #http://wiki.wireshark.org/Gratuitous_ARP
    cmd = ["arping", "-q", "-U", "-c", "2", "-I", dev_ip, vip]
    run_cmd(cmd)
    cmd = ["arping", "-q", "-A", "-c", "2", "-I", dev_ip, vip]
    run_cmd(cmd)


def ifconfig_vip_down(vip):
    """Release vip."""
    dev = _get_ip_dev(vip)
    if dev:
        cmd = ["ifconfig", dev, "down"]
        run_cmd(cmd)


def is_holding_vip(vip):
    """Check whether the current server is holding the vip."""
    dev = _get_ip_dev(vip)
    return dev and True or False


def validate_vip_binding(vip, ip):
    """Check if vip is correctly bound to ip."""
    dev_vip = _get_ip_dev(vip)
    dev_ip = _get_ip_dev(ip)
    return (dev_vip != dev_ip and ':' in dev_vip and
            dev_vip.split(':')[0] == dev_ip.split(':')[0])


def iscsi_logout_target(target, portal):
    cmd = ["/sbin/iscsiadm",
           "--mode", "node",
           "--targetname=%s" % target,
           "--portal=%s" % portal,
           "--logout"]
    run_cmd(cmd)


def iscsi_delete_portal(portal):
    cmd = ["/sbin/iscsiadm",
           "--mode", "node",
           "--portal=%s" % portal,
           "--op=delete"],
    run_cmd(cmd)


def get_file_size(filename):
    """
    Return the size of a file. If the file is a block device then
    this returns the size of the block device.

    """
    file_stat = os.stat(filename)
    file_size = 0
    if stat.S_ISREG(file_stat[stat.ST_MODE]):
        file_size = file_stat[stat.ST_SIZE]
    elif stat.S_ISBLK(file_stat[stat.ST_MODE]):
        file_size = run_cmd(['/sbin/blockdev', '--getsize64', filename])

    return file_size


def get_copy_cmd(src, dst, sparse):
    """
    Return the appropriate command to copy src to dst.

    """
    args = ['/bin/dd', 'if=%s' % src, 'of=%s' % dst, 'bs=1M',
            'iflag=direct', 'oflag=direct']
    if sparse:
        args.append('conv=sparse')
    return args


def sync():
    """
    Synchronize data on disk with memory

    """
    run_cmd(['sync'])


def get_fs_type(fs_location):
    if fs_location.startswith('/dev/'):
        return 'ocfs2'
    else:
        return 'nfs'


def do_export_nfs(client, path, options):
    if not os.path.ismount(path):
        logging.warn('%s is not a mount point', path)
    if not options:
        options = 'rw,async,no_root_squash'
    cmd = ['service', 'nfs', 'start']
    run_cmd(cmd)
    client_path = '%s:%s' % (client, path)
    cmd = ['exportfs', '-o', options, client_path]
    run_cmd(cmd)


def do_unexport_nfs(client, path):
    client_path = '%s:%s' % (client, path.rstrip('/'))
    cmd = ['exportfs', '-u', client_path]
    run_cmd(cmd)
    if not run_cmd(['exportfs']):
        cmd = ['service', 'nfs', 'stop']
        run_cmd(cmd)


def mount(resource, mountpoint, fstype=None):
    """
    Mount a file system.

    Mount a file system. The mount point is created if it does not
    exist. This does nothing if the file system is already mounted.
    If the mount fails then the mount point is removed and an
    exception is raised.

    @param resource: The resource to mount.
    @type  resource: C{string}

    @param mountpoint: The mount point where the resource has to be
    mounted.
    @type  mountpoint: C{string}

    @param fstype: The type of file system to mount.
    @type  fstype: C{string}

    """
    if os.path.ismount(mountpoint):
        return

    if not os.path.exists(mountpoint):
        os.makedirs(mountpoint)

    cmd = ['mount']
    if fstype:
        cmd += ['-t', fstype]

    cmd += [resource, mountpoint]

    try:
        run_cmd(cmd)
    except Exception, err:
        try:
            os.rmdir(mountpoint)
        except Exception, e:
            logging.warn('Error removing mount point: %s', mountpoint)
        raise err


def umount(mountpoint):
    """
    Unmount a file system.

    Unmount a file system. The mount point is removed after the file
    system has been successfully unmounted. This does nothing if the
    file system is not mounted.

    @param mountpoint: The mount point that should be unmounted.
    @type  mountpoint: C{string}

    """
    if not os.path.ismount(mountpoint):
        return

    run_cmd(['umount', mountpoint])
    os.rmdir(mountpoint)
