# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import re
import time
import xen.lowlevel.xc
import xen.xend.XendClient
from xen.xend.xenstore.xstransact import xstransact
from agent.target.host import run_cmd
from agent.target.hypervisor import get_vms
from agent.lib.stats import StatsCollectorBase


def read_file(*args):
    tmpfile = open(os.path.join(*args))
    try:
        return tmpfile.read().strip()
    finally:
        tmpfile.close()


def read_xenstore(path):
    try:
        return xstransact.Read(path).strip()
    except Exception, err:
        logging.debug('Error reading xenstore (%s): %s', path, err)


def get_vm_stats():
    try:
        return get_vms()
    except Exception, err:
        logging.debug('Error getting VM stats: %s', err)


def get_disk_uuid(dev):
    try:
        info = run_cmd(['dmsetup', 'info', '--columns', '--noheadings',
                        '--options=UUID', '/dev/%s' % dev])
        if not info.startswith('part'):
            return info.rstrip('\n').replace('mpath-', '')
    except Exception:
        # dmnfs will not create disk nodes under /dev, ignore them.
        pass


def get_disk_stat():
    disk_read = {}
    disk_write = {}
    try:
        f = open('/proc/diskstats')
        try:
            for line in f:
                stat = line.split()
                if not stat[2].startswith('dm-'):
                    continue
                disk_uuid = get_disk_uuid(stat[2])
                if not disk_uuid:
                    continue
                read = long(stat[5]) << 9
                write = long(stat[9]) << 9
                disk_read[disk_uuid] = read
                disk_write[disk_uuid] = write
        finally:
            f.close()
    except Exception, err:
        logging.debug('Error getting Server disk read/write: %s', err)
    return disk_read, disk_write


def get_port_stat():
    port_rx = {}
    port_tx = {}
    try:
        basedir = '/sys/class/net'
        for entry in os.listdir(basedir):
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            if entry.startswith('vif') or entry == 'lo':
                continue
            rx = read_file(statdir, 'rx_bytes')
            tx = read_file(statdir, 'tx_bytes')
            port_rx[entry] = long(rx)
            port_tx[entry] = long(tx)
    except Exception, err:
        logging.debug('Error getting Server port RX/TX: %s', err)
    return port_rx, port_tx


def get_vm_disk_stat(domid):
    disk_read = {}
    disk_write = {}
    try:
        basedir = '/sys/bus/xen-backend/devices'
        for entry in os.listdir(basedir):
            if not entry.startswith('vbd-%s' % domid):
                continue
            nodename = read_file(basedir, entry, 'nodename')
            backend = read_xenstore('/local/domain/0/%s/params' % nodename)
            if not backend:
                continue
            # remove leading 'aio:', 'ioemu:', etc.
            backend = backend.split(':', 1)[-1]
            read = read_file(basedir, entry, 'statistics/rd_sect')
            write = read_file(basedir, entry, 'statistics/wr_sect')
            disk_read[backend] = long(read)
            disk_write[backend] = long(write)
    except Exception, err:
        logging.debug('Error getting VM disk read/write: %s', err)
    return disk_read, disk_write


def get_vm_port_stat():
    # FIXME: should we get vif statistics for each VM?
    port_rx = {}
    port_tx = {}
    try:
        basedir = '/sys/class/net'
        for entry in os.listdir(basedir):
            if not re.match(r'^vif\d+\.\d+$', entry):
                continue
            statdir = os.path.join(basedir, entry, 'statistics')
            if not os.path.exists(statdir):
                continue
            # backend TX == VM frontend RX
            # backend RX == VM frontend TX
            rx = read_file(statdir, 'tx_bytes')
            tx = read_file(statdir, 'rx_bytes')
            port_rx[entry] = long(rx)
            port_tx[entry] = long(tx)
    except Exception, err:
        logging.debug('Error getting VM port RX/TX: %s', err)
    return port_rx, port_tx


def tmem_parse(line):
    parsed = {}
    for item in line[2:].split(','):
        key, value = item.split(':')
        if key in ['U0', 'U1']:
            parsed[key] = int(value, 16)
        elif key == 'PT':
            parsed[key] = value
        else:
            parsed[key] = int(value)
    return parsed


def tmem_list():
    xc = xen.lowlevel.xc.xc()
    raw = xc.tmem_control(-1, 4, -1, 32768, 1, 0, '')
    if raw == -1:
        return {}
    parsed = {'global': {},
              'time_stats': {},
              'clients': [],
              'pools': [],
              'shared_pools': []}
    for line in raw.splitlines():
        if line.startswith('G='):
            parsed['global'] = tmem_parse(line)
        if line.startswith('T='):
            parsed['time_stats'] = tmem_parse(line)
        if line.startswith('C='):
            parsed['clients'].append(tmem_parse(line))
        if line.startswith('P='):
            parsed['pools'].append(tmem_parse(line))
        if line.startswith('S='):
            parsed['shared_pools'].append(tmem_parse(line))
    result = {}
    for pool in parsed['pools']:
        client_id = pool['CI']
        if client_id not in result:
            result[client_id] = [0, 0]
        pgp_count = pool['Pc']
        if pool['PT'] == 'EP':
            result[client_id][0] += pgp_count
        if pool['PT'] == 'PP':
            result[client_id][1] += pgp_count
    return result


def tmem_query_freeable_mb():
    xc = xen.lowlevel.xc.xc()
    freeable_mb = xc.tmem_control(-1, 8, -1, 0, 0, 0, '')
    if freeable_mb == -1:
        return 0
    else:
        return freeable_mb


class StatsCollector(StatsCollectorBase):
    def __init__(self, queue, lock):
        StatsCollectorBase.__init__(self, queue, lock)
        self._previous_timestamp = 0.0
        self._previous_cpu_time = {}

    def _run(self):
        vm_stats = get_vm_stats()
        if not vm_stats:
            return
        nodeinfo = dict(xen.xend.XendClient.server.xend.node.info())
        tmeminfo = tmem_list()
        current_timestamp = time.time()
        nr_cpus = nodeinfo['nr_cpus']

        disk_read, disk_write = get_disk_stat()
        port_rx, port_tx = get_port_stat()
        vm_port_rx, vm_port_tx = get_vm_port_stat()

        ptmem = sum([x for _, x in tmeminfo.values()])
        etmem = tmem_query_freeable_mb()

        total_cpu_load = 0.0
        current_cpu_time = {}
        for vm in vm_stats:
            try:
                domid = vm['domid']
                name = vm['name']
                vcpu_count = int(vm['vcpus'])
                target_mem = vm['memory']

                xc = xen.lowlevel.xc.xc()
                try:
                    xcdominfo = xc.domain_getinfo(domid, 1)
                    current_mem = xcdominfo[0]['mem_kb'] / 1024
                except Exception:
                    current_mem = target_mem

                if domid in tmeminfo:
                    vmetmem = tmeminfo[domid][0]
                    vmptmem = tmeminfo[domid][1]
                else:
                    vmetmem = 0
                    vmptmem = 0

                vm_disk_read, vm_disk_write = get_vm_disk_stat(domid)

                cpu_util = 0.0
                cpu_load = 0.0
                current_vm_cpu_time = float(vm['cpu_time'])
                previous_vm_cpu_time = self._previous_cpu_time.get(vm['uuid'])
                if previous_vm_cpu_time:
                    time_delta = current_timestamp - self._previous_timestamp
                    if current_vm_cpu_time > previous_vm_cpu_time:
                        cpu_time_delta = current_vm_cpu_time - previous_vm_cpu_time
                    else:
                        cpu_time_delta = current_vm_cpu_time
                    cpu_load = cpu_time_delta * 100 / time_delta
                    cpu_load_ceiling = float(vcpu_count * 100)
                    if cpu_load > cpu_load_ceiling:
                        cpu_load = cpu_load_ceiling
                    cpu_util = float(cpu_load / cpu_load_ceiling)
                    total_cpu_load += cpu_load

                # FIXME: some items are added for backward compatibility:
                #   vmDomid
                #   vmCpuLoad
                self.queue_stat('vmDomid', int(domid), name)
                self.queue_stat('vmCurrentMemory', long(current_mem), name)
                self.queue_stat('vmTargetMemory', long(target_mem), name)
                self.queue_stat('vmMaxMemory', long(vm['maxmem']), name)
                self.queue_stat('vmEphemeralTranscendentMemory', long(vmetmem), name)
                self.queue_stat('vmPersistentTranscendentMemory', long(vmptmem), name)
                self.queue_stat('vmOnlineVcpus', int(vm['online_vcpus']), name)
                self.queue_stat('vmCpuTime', current_vm_cpu_time, name)
                self.queue_stat('vmCpuLoad', cpu_load, name)
                self.queue_stat('vmCpuUtil', cpu_util, name)
                self.queue_stat('vmDiskRead', vm_disk_read, name)
                self.queue_stat('vmDiskWrite', vm_disk_write, name)

                current_cpu_time[vm['uuid']] = current_vm_cpu_time
            except Exception, err:
                logging.warn('Error getting VM statistics: %s', err)
                continue

        total_cpu_load_ceiling = float(nr_cpus * 100)
        if total_cpu_load > total_cpu_load_ceiling:
            total_cpu_load = total_cpu_load_ceiling
        total_cpu_util = float(total_cpu_load / total_cpu_load_ceiling)
        total_memory = long(nodeinfo['total_memory'])
        free_memory = long(nodeinfo['free_memory'])
        used_memory = total_memory - free_memory

        # FIXME: some items are added for backward compatibility:
        #   vmDom0FreeMemory
        #   vmDom0TotalMemory
        #   vmDom0TotalCpuUtil
        self.queue_stat('serverUsedMemory', used_memory)
        self.queue_stat('vmDom0TotalMemory', total_memory)
        self.queue_stat('serverFreeMemory', free_memory)
        self.queue_stat('vmDom0FreeMemory', free_memory)
        self.queue_stat('serverEphemeralTranscendentMemory', long(etmem))
        self.queue_stat('serverPersistentTranscendentMemory', long(ptmem))
        self.queue_stat('serverCpuUtil', total_cpu_util)
        self.queue_stat('vmDom0TotalCpuUtil', total_cpu_util)
        self.queue_stat('serverPortReceive', port_rx)
        self.queue_stat('serverPortSent', port_tx)
        self.queue_stat('serverDiskRead', disk_read)
        self.queue_stat('serverDiskWrite', disk_write)
        self.queue_stat('vmNetworkReceive', vm_port_rx)
        self.queue_stat('vmNetworkWrite', vm_port_tx)

        # FIXME: below are junk statistics and should be removed.
        meminfo = dict([item.strip() for item in line.split(':', 1)]
                       for line in read_file('/proc/meminfo').splitlines())
        self.queue_stat('serverFreeMem', meminfo['MemFree'])
        self.queue_stat('serverFreeSwap', meminfo['SwapFree'])
        server_cpu_load = float(read_file('/proc/loadavg').split()[0])
        self.queue_stat('serverCpuLoad', server_cpu_load)

        self._previous_timestamp = current_timestamp
        self._previous_cpu_time = current_cpu_time
