# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import copy
import logging
import signal
from agent.target.api import MODULES


TRACE_EXCEPTION = [
    'dispatch',
    'ovs_async_proc',
    'get_last_boot_time',
    'update_agent_password',
    'send_to_guest',
    'list_vms',
    'remote_power_on',
    'remote_power_off',
]

def well_behavior(decorator):
    """This decorator can be used to turn simple functions
    into well-behaved decorators, so long as the decorators
    are fairly simple. If a decorator expects a function and
    returns a function (no descriptors), and if it doesn't
    modify function attributes or docstring, then it is
    eligible to use this. Simply apply @well_behavior to
    your decorator and it will automatically preserve the
    docstring and function attributes of functions to which
    it is applied."""
    def new_decorator(f):
        g = decorator(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g
    # Now a few lines needed to make well_behavior itself
    # be a well-behaved decorator.
    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)
    return new_decorator


def scrub_dict(d):
    if isinstance(d, dict):
    	for key, value in d.items():
	    if isinstance(key, (list, tuple, dict)):
		scrub_dict(key)
	    elif isinstance(key, str) and ('passwd' in key or 'password' in key):
	        d[key] = '******'
	    if isinstance(value, (list, tuple, dict)):
	    	scrub_dict(value)
    elif isinstance(d, (list, tuple)):
        for item in d:
            scrub_dict(item)


def scrub_and_log(msg, name, *args):
    if name not in TRACE_EXCEPTION:
        temp_args = copy.deepcopy(args)
	for entry in temp_args:
	    scrub_dict(entry)
        if temp_args:
            logging.debug("%s: %s%s", msg, name, temp_args)
        else:
            logging.debug("%s: %s", msg, name)


@well_behavior
def trace(func):
    def wrapper(*args):
        name = func.__name__
        scrub_and_log('call start', name, *args)
        result = func(*args)
        scrub_and_log('call complete', name)
        return result
    return wrapper


@well_behavior
def catch_error(func):
    def wrapper(*args):
        try:
            return func(*args)
        except Exception, err:
            logging.exception("catch_error: %s", err)
            raise
    return wrapper


def default_stop_handler(sig, frame):
    raise Exception('Task stopped by user.')


@well_behavior
def handle_stop(func):
    def wrapper(*args):
        signal.signal(signal.SIGUSR2, default_stop_handler)
        return func(*args)
    return wrapper


def expose(name):
    @well_behavior
    def wrapper(func):
        def f(*args):
            new_func = catch_error(func)
            new_func = trace(new_func)
            return new_func(*args)
        f.__name__ = name
        return f
    return wrapper


def get_api(version=None):
    api = {}
    for mod in MODULES:
        services = mod().get_services(version)
        for (name, func) in services.items():
            api.update({name: expose(name)(func)})
    return api


def get_api_function(api):
    for mod in MODULES:
        services = mod().get_services()
        for name, func in services.items():
            if name == api:
                return handle_stop(func)
    raise RuntimeError('No such API: %s' % api)
