# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
from distutils.sysconfig import get_python_lib

from agent.lib.config import get_config
from agent.lib.constants import (LOCAL_SERVER_DB,
                                 MEMBERSHIP_STATE_CLUSTERED,
                                 MEMBERSHIP_STATE_HB_CONFIGURED,
                                 MEMBERSHIP_STATE_POOLED,
                                 MEMBERSHIP_STATE_OWNED,
                                 MEMBERSHIP_STATE_UNOWNED,
                                 OVS_REPO_HEAD)
from agent.lib.db import read_item, get_cluster_db_home
from agent.target.info import API_VERSION


def get_api_version():
    return API_VERSION


def get_core_api_version():
    try:
        f = open(os.path.join(get_python_lib(), 'version.properties'))
        d = {}
        for line in f:
            if '=' not in line:
                continue
            (k, v) = line.split('=', 1)
            k = k.strip()
            v = v.strip()
            d[k] = v
        f.close()
        return '.'.join([d['major.version'], d['minor.version'],
                         d['patch.version'], d['bld.version']])
    except Exception, err:
        logging.warn('cannot get core api version: %s', err)


def get_agt_url(host):
    config = get_config()
    if config.get("security", "ssl").lower().startswith("enable"):
        protocol = "https"
    else:
        protocol = "http"
    return "%s://%s:%s" % (protocol, host, config.get("agent", "port"))


def get_repo_path(repo_id):
    return os.path.join(OVS_REPO_HEAD, repo_id)


def get_assembly_path(repo_id, assembly_id):
    return os.path.join(get_repo_path(repo_id), 'Assemblies', assembly_id)


def get_iso_path(repo_id, iso_id):
    return os.path.join(get_repo_path(repo_id), 'ISOs', iso_id)


def get_template_path(repo_id, template_id):
    return os.path.join(get_repo_path(repo_id), 'Templates', template_id)


def get_virtualdisk_path(repo_id, virtualdisk_id):
    return os.path.join(get_repo_path(repo_id), 'VirtualDisks', virtualdisk_id)


def get_vm_path(repo_id, vm_id):
    return os.path.join(get_repo_path(repo_id), 'VirtualMachines', vm_id)


def get_vm_lock_file(repo_id, lock_id):
    return '/var/run/ovs-agent/vm-%s-%s.lock' % (repo_id, lock_id)


def get_cluster_flag():
    return read_item(LOCAL_SERVER_DB, "clustered")


def get_owning_manager_uuid():
    return read_item(LOCAL_SERVER_DB, "manager_uuid")


def get_poolfs_uuid():
    return read_item(LOCAL_SERVER_DB, "poolfs_uuid")


def get_server_pool_uuid():
    return read_item(LOCAL_SERVER_DB, "pool_uuid")


def get_membership_state():
    """
    Return the current state of the server with respect to
    server pool and cluster membership.

    @return     One of: "Unowned", "Owned", "Pooled",
                "Heartbeat_Configured" and "Clustered".
    """
    if get_cluster_flag():
        return MEMBERSHIP_STATE_CLUSTERED

    if get_poolfs_uuid():
        return MEMBERSHIP_STATE_HB_CONFIGURED

    if get_server_pool_uuid():
        return MEMBERSHIP_STATE_POOLED

    if get_owning_manager_uuid():
        return MEMBERSHIP_STATE_OWNED

    return MEMBERSHIP_STATE_UNOWNED


def is_master():
    """Check whether the current server is the server pool master."""
    my_hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
    if not my_hostname:
        return False
    master_hostname = read_item("server_pool", "pool_master_hostname",
                                get_cluster_db_home())
    return my_hostname == master_hostname
