# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import threading
import time
from agent.lib.constants import LOCAL_SERVER_DB
from agent.lib.db import write_item, read_item
from agent.lib.notification import notify_manager
from agent.target.host import get_server_unique_id


STAT_INTERVAL_DEFAULT = 20


def set_statistic_interval(interval):
    write_item(LOCAL_SERVER_DB, "stat_interval", interval)


def get_statistic_interval():
    return read_item(LOCAL_SERVER_DB, "stat_interval") or STAT_INTERVAL_DEFAULT


def sleep_for_statistic_interval():
    elapse = 0
    while True:
        time.sleep(1)
        elapse += 1
        statistic_interval = get_statistic_interval()
        if statistic_interval == -1:
            continue
        if elapse >= statistic_interval:
            break


def get_max_statistic_entries():
    # We try to retain 30 minutes statistics, but we don't want the store too
    # many statistics in memory. When statistic_interval < 5, we use 5 to
    # calculate the queue size.
    # On every collection, we get this number of stats (11 items per VM, and 17
    # items per Server):
    #   11 * number-of-vms + 17
    # We assume each Server has 20 VMs by average.
    statistic_interval = get_statistic_interval()
    if statistic_interval == -1:
        statistic_interval = STAT_INTERVAL_DEFAULT
    if statistic_interval < 5:
        statistic_interval = 5
    return (30 * 60 / statistic_interval or 1) * (11 * 20 + 17)


def process_stats(queue, lock):
    lock.acquire()
    try:
        if not queue:
            return
        stats = {
            'machine': get_server_unique_id(),
            'pool': read_item(LOCAL_SERVER_DB, 'pool_uuid') or '',
            'serialization_version': 1L,
            'stats': list(queue),
        }
        try:
            # FIXME: backward compatibility: Manager requires a list of stats.
            notify_manager('{STATS}', [stats])
        except Exception:
            return
        queue.clear()
    finally:
        lock.release()


class StatsCollectorBase(threading.Thread):
    def __init__(self, queue, lock):
        threading.Thread.__init__(self)
        self.queue = queue
        self.lock = lock
        self._stop = False

    def run(self):
        while not self._stop:
            try:
                self._run()
            except Exception, err:
                logging.exception('Error collecting stats: %s', err)
            sleep_for_statistic_interval()

    def _run(self):
        # to be overridden by subclasses
        pass

    def stop(self):
        self._stop = True

    def queue_stat(self, stype, value, name=None):
        #FIXME: 'id' is meaningless, should we remove or redefine it?
        ident = 'localhost'
        if name:
            ident = 'localhost/%s' % name
        stat = {
            'id': ident,
            'timestamp': long(time.time() * 1000),
            'type': stype,
            'value': value,
            }
        self.lock.acquire()
        try:
            maxlen = get_max_statistic_entries()
            while len(self.queue) >= maxlen:
                self.queue.popleft()
            self.queue.append(stat)
        finally:
            self.lock.release()


if __name__ == '__main__':
    # for debugging only
    # usage: python stats.py [stat_type] ...
    import collections
    import pprint
    import sys
    from agent.target.stats import StatsCollector
    logging.basicConfig(level=logging.DEBUG)
    stypes = sys.argv[1:]
    queue = collections.deque()
    lock = threading.Lock()
    collector = StatsCollector(queue, lock)
    collector.start()
    while True:
        try:
            sleep_for_statistic_interval()
            lock.acquire()
            try:
                if not queue:
                    continue
                if stypes:
                    for stat in list(queue):
                        if stat['type'] in stypes:
                            pprint.pprint(stat)
                else:
                    pprint.pprint(list(queue))
                queue.clear()
            finally:
                lock.release()
        except Exception, err:
            collector.stop()
            collector.join()
            logging.exception(err)
            sys.exit(1)
