# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os

from agent.lib.settings import get_vm_path


def create_cfg(path, params):
    check_cfg(params)
    try:
        write_cfg(path, params)
    except Exception:
        os.remove(path)
        raise


def cfg_exists(path):
    return os.path.exists(path)


def get_cfg_path(repo_id, vm_id):
    return get_cfg_file(get_vm_path(repo_id, vm_id))


def get_cfg_file(path):
    return os.path.join(path, 'vm.cfg')


def read_cfg(path):
    cfg = {}
    execfile(path, {}, cfg)
    return cfg


def write_cfg(path, cfg):
    fix_cfg(cfg)
    check_cfg(cfg)
    _write_vmcfg(path, cfg)


def _write_vmcfg(path, cfg):
    f = open(path, 'w')
    for (k, v) in cfg.items():
        f.write('%s = %s\n' % (k, repr(v)))
    f.close()


def configure_cfg(path, params):
    check_cfg(params)
    for option in sorted(params.keys()):
        value = params[option]
        try:
            update_cfg(path, option, value)
        except Exception, err:
            raise RuntimeError('%s:config_error:%s' % (option, err))


def unconfigure_cfg(path, params):
    cfg = read_cfg(path)
    changed = False
    for option in params:
        if option in cfg:
            del cfg[option]
            changed = True
    if changed:
        write_cfg(path, cfg)


def update_cfg(path, option, value):
    cfg = read_cfg(path)
    if option not in cfg or cfg[option] != value:
        cfg[option] = value
        write_cfg(path, cfg)


def check_cfg(cfg):
    # FIXME: Are we going to support arbitrary options? Yes for now.
    for (option, value) in cfg.items():
        if option in VM_CFG:
            checker = VM_CFG[option]
            try:
                checker(value)
            except Exception, err:
                raise RuntimeError('%s:type_error:%s' % (option, err))


def fix_cfg(cfg):
    for (option, value) in cfg.items():
        if option in VM_CFG:
            checker = VM_CFG[option]
            if checker == check_int:
                try:
                    cfg[option] = int(value)
                except ValueError, err:
                    raise RuntimeError('Cannot fix option (%s) value (%s): %s' %
                                       (option, value, err))
def check_str(value):
    if not isinstance(value, str):
        raise RuntimeError('A string is required, but %s is provided.' %
                           type(value))


def check_int(value):
    try:
        int(value)
    except ValueError, err:
        raise RuntimeError('An integer is required: %s.' % err)


def check_list(value):
    if not isinstance(value, list):
        raise RuntimeError('A list is required, but %s is provided.' %
                           type(value))


def check_str_or_list(value):
    if not isinstance(value, (str, list)):
        raise RuntimeError('A string or list is required, ' +
                           'but %s is provided.' % type(value))


VM_CFG =  {
    # option: checker,
    'access_control': check_list,
    'acpi': check_int,
    'apic': check_int,
    'blkif': check_int,
    'bootargs': check_str,
    'bootloader': check_str,
    'boot': check_str,
    'builder': check_str,
    'console_autoconnect': check_int,
    'cpu_cap': check_int,
    'cpuid_check': check_list,
    'cpuid': check_list,
    'cpus': check_str_or_list,
    'cpu': check_int,
    'cpu_weight': check_int,
    'device_model': check_str,
    'dhcp': check_str,
    'disk_other_config': check_list,
    'disk': check_list,
    'display': check_int,
    'expose_host_uuid': check_int,
    'extra': check_str,
    'fda': check_str,
    'fdb': check_str,
    'features': check_str,
    'gateway': check_str,
    'guest_os_type': check_str,
    'hap': check_int,
    'hostname': check_str,
    'hpet': check_int,
    'interface': check_str,
    'ioports': check_list,
    'ip': check_str,
    'irq': check_list,
    'isa': check_int,
    'kernel': check_str,
    'keymap': check_str,
    'loader': check_str,
    'localtime': check_int,
    'machine_address_size': check_int,
    'maxmem': check_int,
    'maxvcpus': check_int,
    'memory': check_int,
    'monitor_path': check_str,
    'monitor': check_int,
    'name': check_str,
    'netif': check_int,
    'netmask': check_str,
    'nfs_root': check_str,
    'nfs_server': check_str,
    'nographic': check_int,
    'nomigrate': check_int,
    'on_crash': check_str,
    'on_poweroff': check_str,
    'on_reboot': check_str,
    'on_xend_start': check_str,
    'on_xend_stop': check_str,
    'opengl': check_int,
    'pae': check_int,
    'pasued': check_int,
    'pci_msitranslate': check_int,
    'pci_power_mgmt': check_int,
    'pci': check_list,
    'ramdisk': check_str,
    'root': check_str,
    'rtc_timeoffset': check_int,
    's3_integrity': check_int,
    'sdl': check_int,
    'serial': check_str,
    'shadow_memory': check_int,
    'soundhw': check_str,
    'stdvga': check_int,
    'superpages': check_int,
    'suppress_spurious_page_faults': check_int,
    'target': check_int,
    'timer_mode': check_int,
    'tpmif': check_int,
    'tsc_mode': check_int,
    'usbdevice': check_str,
    'usb': check_int,
    'uuid': check_str,
    'vcpus': check_int,
    'vfb': check_list,
    'vhpt': check_int,
    'videoram': check_int,
    'vif2': check_list,
    'vif_other_config': check_list,
    'vif': check_list,
    'viridian': check_int,
    'vncconsole': check_int,
    'vncdisplay': check_int,
    'vnclisten': check_str,
    'vncpasswd': check_str,
    'vncunused': check_int,
    'vnc': check_int,
    'vpt_align': check_int,
    'vscsi': check_list,
    'vtpm': check_list,
    'xauthority': check_str,
    'xen_platform_pci': check_int,
}
