#!/usr/bin/env python
#
# Copyright (C) 2011-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Konrad Rzeszutek
#
"""Utilities for dealing with SAS SAN devices.
"""

import st_utils
import st_os

import os
import re

from blockDeviceOperations	import *

####################################################################################################

__all__ = [	'SAS_device_state_OBJ'
		,	'get_SAS_end_device_list'
		,	'get_SAS_blk_dict']

####################################################################################################

class SAS_device_state_OBJ:

	def __init__( self ):
		self.mSCSIHostNum   = -1
		self.mSCSIId		= -1
		self.mPortId		= -1
		self.mPHY          	= -1
		self.mDeviceAddress	= None	# The end-point. In FC the node_name (Target WWPN)
		self.mHBAAddress	= None  # The initiator. In FC the port_name (HBA's WWPN)
		self.mBlkDev_LIST	= []
		self.mSGDev_LIST	= []

	def __repr__( self ):
		return "mSCSIHostNum: %d mSCSIId: %d mPortId: %d mPHY: %d mHBAAddress: %s mBlkDevs: %s mSGDevs: %s" % (
						self.mSCSIHostNum
					,	self.mSCSIId
					,	self.mPortId
					,	self.mPHY
					,	self.mHBAAddress
					,	self.mBlkDev_LIST
					,	self.mSGDev_LIST)

	def __str__ (self) :
		return "%d:%d:%d,%d %s->%s" % (self.mSCSIHostNum, self.mSCSIId, self.mPortId, self.mPHY, self.mHBAAddress, self.mDeviceAddress)

	def __cmp__( self, other):
		if not other:
			return -1

		if self.mSCSIHostNum < other.mSCSIHostNum:
			return -1
		if self.mSCSIId < other.mSCSIId:
			return -1
		if self.mPortId < other.mPortId:
			return -1
 		if self.mPHY < other.mPHY:
			return -1
 		if self.mHBAAddress < other.mHBAAddress:
			return -1

		if self.mSCSIHostNum > other.mSCSIHostNum:
			return 1
		if self.mSCSIId > other.mSCSIId:
			return 1
		if self.mPortId > other.mPortId:
			return 1
 		if self.mPHY > other.mPHY:
			return 1
 		if self.mHBAAddress > other.mHBAAddress:
			return 1

		return 0

####################################################################################################
class SAS_SAN_BlkDev_Operations(BlkDev_Operations):

	"""SAS_SAN class for BlkDev_Operations
	"""

	########################################
	def __init__( self, blkDevInfo, SAS_device_state ):

		self.mBlkDevInfo		= blkDevInfo
		self.mSAS_device	= SAS_device_state

	########################################
	def __repr__( self ):

		return "SAS_SAN_BlkDev_Operations[%s]" % ( self.mBlkDevInfo.mBlkDev )

	########################################
	def disconnect_block_device( self, vi_errorAndLog_obj, forceDisconnection ):
		"""
		"""
	#
		std_disconnect_block_device( vi_errorAndLog_obj, self.mBlkDevInfo, forceDisconnection )
		return 0

	########################################
	def revalidate_block_device( self, vi_errorAndLog_obj ):
		"""
		Updates a standard block device to the latests sysfs values
		"""
	#
		retVal	= std_revalidate_block_device( vi_errorAndLog_obj, self.mBlkDevInfo )
		if retVal:
			return retVal

		return 0


####################################################################################################
def get_SAS_end_device_list( vi_errorAndLog_obj, refresh ):
	"""
	This routine extracts the SAS SAN data. The returned values are in the
	list wherein the values are SAS_device_state_OBJ objects.
	"""

	SAS_SAN_end_device_LIST = []
	#
	# Directory format of ../sas_device is as follow:
	#	end_device-<Z>:<A>:<Y>
	#	end_device-<Z>:<Y> (Y is Port Identifier)
	#		-phy_identifier
	#		-sas_address
	#		+ device
	#			+target<Z>:0:0
	#				+ Z:A:B:C (the SCSI Host:Bus:Channel:LUN ID)
	#					+ block:<sdM> (symlink to /sys/block/sdM) - optional
	#					+ scsi_generic:<sgN> (symlink to /sys/scsi_generic/sgN)
	#					+ and more dirs...
	#					- delete, type, vendor, model, rescan, etc..
	#

	sas_device_root = '/sys/class/sas_device/'
	if not os.path.exists( sas_device_root ):
		return []

	for sas_device in os.listdir( sas_device_root ):
		# We only want the end_device-<SCSI Host>:<SCSI ID>:<Port Identifier>, the end_device-<SCSI Host>:<Port Identifier>
		# is for non-SAS SAN setups.
		scsi_is_sas_expander_device_match = re.match("^\S*end_device-(\d+):(\d+):(\d+)$", sas_device)
		if scsi_is_sas_expander_device_match:

			SAS_device_state             	= SAS_device_state_OBJ()
			SAS_device_state.mSCSIHostNum 	= int(scsi_is_sas_expander_device_match.group(1))
			SAS_device_state.mSCSIId 		= int(scsi_is_sas_expander_device_match.group(2))
			SAS_device_state.mPortId 		= int(scsi_is_sas_expander_device_match.group(3))

			mPHY,read_failed 		= st_utils.vi_read_text_value( vi_errorAndLog_obj, sas_device_root+ sas_device + "/phy_identifier" )
			if not read_failed:
				SAS_device_state.mPHY	= int(mPHY)

			SAS_device_state.mHBAAddress, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj,sas_device_root + sas_device + "/sas_address" )
			SAS_device_state.mDeviceAddress, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, sas_device_root + sas_device + "/enclosure_identifier" )

			device_root = "/sys/class/sas_device/%s/device" % (sas_device)
			for root, dirs, files in os.walk( device_root ):
				for dir in dirs:
					# Extract the block device and SCSI Generic name
					if 'target' in dir:
						# Silly sym-links means that we have to re-do a walk on the .../target<Z>:0:0 directory:
						target_root = sas_device_root + "%s/device/%s/" % (sas_device, dir)
						for root, dirs, files in os.walk(target_root):
							# 12:0:0:1, 12:0:0:3, etc..
							for dir in dirs:
								if 'scsi_generic:' in dir:
									sgname = re.sub( 'scsi_generic:', '', dir)
									SAS_device_state.mSGDev_LIST.append(sgname)
								if 'block:' in dir:
									blkName = re.sub( 'block:', '', dir)
									SAS_device_state.mBlkDev_LIST.append(blkName)

			SAS_SAN_end_device_LIST.append( SAS_device_state )
			#vi_errorAndLog_obj.vi_print_dbg( 0, SAS_device_state )

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, SAS_SAN_host_DICT )

	return SAS_SAN_end_device_LIST

####################################################################################################
def get_SAS_blk_dict( vi_errorAndLog_obj, refresh ):
	"""
	This routine creates a dictionary wherein the key is the block device name ("sda") and the
	value is the associated SAS_device_state_OBJ.
	"""

	SAS_SAN_end_device_LIST = get_SAS_end_device_list(vi_errorAndLog_obj, refresh)

	SAS_SAN_blk_DICT = {}

	for SAS_device_state in SAS_SAN_end_device_LIST:
		for blkDev in SAS_device_state.mBlkDev_LIST:
			SAS_SAN_blk_DICT [ blkDev ] = SAS_device_state

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, SAS_SAN_blk_DICT )

	return SAS_SAN_blk_DICT
