#!/usr/bin/env python
#
# Copyright (C) 2011-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Konrad Rzeszutek
#
"""Utilities for dealing with the SCSI Generic device.
"""

import st_utils

import os
import re
import traceback
import errno
import time

from blockDeviceOperations	import *
from scsiDeviceID		import *
from deviceMapper		import *

####################################################################################################

__all__ = [	'SG_device_State_OBJ'
		,	'SG_device_Operations'
		,	'do_sg_inq'
		,	'find_sg_devices'
		,	'find_sg_devices_on_host'
		,	'disconnect_stale_sg_devices_on_host'
		,	'isSysFS_stale'
		,	'sg_device_disconnect'
		,	'sg_device_rescan'
		,	'get_sgdev_to_host_nr'
		,	'get_host_nr_to_sgdev']

####################################################################################################

class SG_device_State_OBJ:

	mSCSI_device_id			= None

	def __init__(self):
		self.mDeviceName		= None
		self.mBlkName			= None	# Optional. If set, this SCSI device has a block device.

		self.mPQual				= -1	# Values obtained from SCSI INQ.
		self.mTGPS				= -1
		self.mRDAC				= -1
		self.mMultiP			= -1

		self.mSCSI_vendor		= None # Values from SCSI INQ cmd.
		self.mSCSI_model		= None
		self.mSCSI_rev			= None
		self.mSCSI_type			= -1
		self.mSCSI_size			= -1
		self.mSysFS_model		= None # vs values from sysfs
		self.mSysFS_vendor		= None
		self.mSysFS_rev			= None
		self.mSysFS_type		= -1
		self.mSysFS_size		= -1

		self.mSCSI_device_id	= None # SCSI_device_ID_OBJ


	def __repr__(self):
		return "mDeviceName: %s mBlkName: %s mPQual: %d mTGPS: %d mRDAC: %d mMultiP: %d (SCSI: mVendor: %s mModel: %s, mRev: %s, mType: %d,mSize: %d;SysFS: mVendor: %s, mModel: %s, mRev: %s, mType: %d, mSize: %d) mSCSI_Dev_ID: %s: " % (
						self.mDeviceName
					,	self.mBlkName
					,	self.mPQual
					,	self.mTGPS
					,	self.mRDAC
					,	self.mMultiP
					,	self.mSCSI_vendor
					,	self.mSCSI_model
					,	self.mSCSI_rev
					,	self.mSCSI_type
					,	self.mSCSI_size
					,	self.mSysFS_vendor
					,	self.mSysFS_model
					,	self.mSysFS_rev
					,	self.mSysFS_type
					,	self.mSysFS_size
					,	self.mSCSI_device_id.as_str() )

	def __str__(self):
		"""
		Output similar to what 'lsscsi' produces.
		"""

		def figure_out_type( scsi_type ):
			"""
			Some comment.
			"""
			device_type_LIST = [ 'disk', 'tape', 'printer', 'process', 'worm', 'cd/dvd', \
							'scanner', 'optical', 'mediumx', 'comms', '(0xa)', '(0xb)', \
							'storage', 'enclosu', 'sim dsk', 'opti rd', 'bridge', 'osd', \
							'adi', '(0x13)', '(0x14)', '(0x15)', '(0x16)', '(0x17)', \
							'(0x18)', '(0x19)', '(0x1a)', '(0x1b)', '(0x1c)', '(0x1e)', \
							'wlun', 'no dev' ]
			if scsi_type > len(device_type_LIST):
				return "unknown"
			return device_type_LIST[scsi_type].strip()

		def if_same_pick_second( first, second ):
			if first == None:
				return second
			if second == None:
				return first
			if first in second:
				return second
			else:
				return "mismatch"

		#[2:0:0:0]    disk    ATA      WDC WD1600JD-75H 08.0  /dev/sda
		if self.mSCSI_device_id:
			str =  "[%s]" % (self.mSCSI_device_id.as_str())
			str =  "%-14s" % (str)
		else:
			str = ''
		# And now the rest of the values.
		if self.mSysFS_type == -1:
			str += " %-7s" % (figure_out_type(self.mSCSI_type))
		else:
			if self.mSCSI_type == self.mSysFS_type:
				str += " %-7s" % (figure_out_type(self.mSCSI_type))
			else:
				str += "mismatch"
		# COMPELNT
		str += " %-8s" % (if_same_pick_second(self.mSysFS_vendor, self.mSCSI_vendor))
		# WDC WD2500SD-01K
		str += " %-16s" % (if_same_pick_second(self.mSysFS_model, self.mSCSI_model))
		str += " %-4s" % (if_same_pick_second(self.mSysFS_rev, self.mSCSI_rev))
		if self.mBlkName:
			str += " /dev/%s" % (self.mBlkName)
		else:
			str += " /dev/%s" % (self.mDeviceName)
		return str

####################################################################################################
def extract_from_sg_inq_data ( sg_device, lines ):
	"""
	Extract from the SCSI INQ command data information.
	"""
#
	#############################################
	def __handle_PQual ( sg_device, val=None):
		if  val:
			sg_device.mPQual = int(val)

	def __handle_Type ( sg_device, val=None):
		if  val:
			sg_device.mSCSI_type = int(val)

	def __handle_TGPS ( sg_device, val=None):
		if  val:
			sg_device.mTGPS = int(val)

	def __handle_MultiP ( sg_device, val=None):
		if  val:
			sg_device.mMultiP = int(val)

	def __handle_Vendor ( sg_device, val=None):
		if  val:
			sg_device.mSCSI_vendor = val.strip()

	def __handle_Model ( sg_device, val=None):
		if val:
			sg_device.mSCSI_model = val.strip()

	def __handle_Rev ( sg_device, val=None):
		if val:
			sg_device.mSCSI_rev = val.strip()

	def __handle_RDAC ( sg_device, val=None):
		if val:
			sg_device.mRDAC = 1

	def __handle_Size ( sg_device, val=None):
		if val:
			if (int(val) >= 512):
				sg_device.mSCSI_size = int(val)/512
			else:
				sg_device.mSCSI_size = 0

	pattern_DICT = { ".*PQual=(\d+).*"			:	__handle_PQual
			,".*Device_type=(\d+).*"			:	__handle_Type
			,".*TGPS=(\d+).*"					:	__handle_TGPS
			,".*MultiP=(\d+).*"					:	__handle_MultiP
			,".*Vendor identification: (.*)"	:	__handle_Vendor
			,".*Product identification: (.*)"	:	__handle_Model
			,".*Product revision level: (.*)"	:	__handle_Rev
			,".*Volume Access Control (RDAC)"	:	__handle_RDAC
			,".*Device size: (\d+) bytes.*"		:	__handle_Size
			}
	# Construct the patterns
	pats_LINE = []
	for pat, handler in pattern_DICT.iteritems():
		pats_LINE.append((re.compile(pat), handler))

	# Start the match.
	for line in lines:
		for pat, handler in pats_LINE:
			m = pat.match(line)
			if m:
				if callable(handler):
					data=m.group(1)
					handler(sg_device, data)

####################################################################################################
def do_sg_inq( vi_errorAndLog_obj, blkDevPath ):
	"""
	Extract the device information from a SCSI Generic INQ command and return a SG_device_State_OBJ
	with that device information.

	Returns:
	SG_device_State_OBJ or None
	"""
#
	stdout_LIST, stderr_LIST, exitStatus, deltaTime	= \
		st_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sg_inq', blkDevPath ] )

	if len(stdout_LIST) < 1:
		return None

	sg_dev_state	= SG_device_State_OBJ()

	extract_from_sg_inq_data ( sg_dev_state, stdout_LIST )

	return sg_dev_state

####################################################################################################
def do_scsi_inq(vi_errorAndLog_obj, sg_dev_state):
	"""
	Extract the device information from a SCSI INQ command and populate the
	SG_device_State_OBJ.

	Returns:
	Zero if successfull.
	"""
#

	retry = 5
	skip_inq = False
	# Retry up to three times.
	while (retry):

			if not skip_inq:
				cmd_str = "sg_inq /dev/%s" % (sg_dev_state.mDeviceName)
				stdout_LIST, stderr_LIST, exitStatus, deltaTime =  \
					st_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_str)

			if skip_inq or (exitStatus == 0 and len(stdout_LIST) > 0):
				extract_from_sg_inq_data ( sg_dev_state, stdout_LIST )
				skip_inq = True

				# Optional. This is to get RDAC information. But we are not using it yet
				if 0:
					cmd_str = "sg_inq -a 0 /dev/%s" % (sg_dev_State.mDeviceName)
					stdout_LIST, stderr_LIST, exitStatus, deltaTime =  \
						st_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_str)
					if len(stdout_LIST) > 0:
						extract_from_sg_inq_data ( sg_dev_state, stdout_LIST )

				cmd_str = "sg_readcap /dev/%s" % (sg_dev_state.mDeviceName)
				stdout_LIST, stderr_LIST, exitStatus, deltaTime =  \
					st_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_str)

				if exitStatus == 0 and len(stdout_LIST) > 0:
					extract_from_sg_inq_data ( sg_dev_state, stdout_LIST )

				if exitStatus == 0:
					return 0

			vi_errorAndLog_obj.vi_print_dbg(2,"Failed to execute %s on %s (%d). Retrying: %d.\nstdout:%s\nstderr:%s" \
				% (cmd_str, sg_dev_state, exitStatus, retry, "".join(stdout_LIST), "".join(stderr_LIST)))
			retry -= 1
			time.sleep ( 1 )

	# CDs,DVDs, etc might fail. Especially as they might have no media.
	if exitStatus:
		vi_errorAndLog_obj.vi_print_dbg(2,"Failed to execute SCSI commands on %s (retried five times). " % (sg_dev_state))

	return exitStatus

####################################################################################################
def do_sysfs_inq(vi_errorAndLog_obj, sg_dev_state):

	"""
	Extract the device information from the SysFS tree and populate the
	SG_device_State_OBJ. The sg_dev_state MUST have the mSCSI_device_id
	object populated.

	Returns:
	Zero if succesfull.
	"""
#
	rc = errno.ENODEV
	sg_dev = sg_dev_state.mSCSI_device_id.as_str()
	device_root = "/sys/class/scsi_device/%s/device/" % (sg_dev)
	for root, dirs, files in os.walk ( device_root ):
		rc = 0
		for dir in dirs:
			# The /dev/sgX value.
			if 'scsi_generic:' in dir:
				sg_dev_state.mDeviceName = re.sub( 'scsi_generic:', '', dir)
			if 'block:' in dir:
				# Aha, this one has a block device associated with it.
				sg_dev_state.mBlkName	= re.sub('block:', '', dir)
				str_size, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, device_root + dir + "/size" )
				if str_size:
					sg_dev_state.mSysFS_size = int(str_size)
		for file in files:
			if 'model' == file:
				sg_dev_state.mSysFS_model, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, device_root + file )
			if 'rev' == file:
				sg_dev_state.mSysFS_rev, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, device_root + file )
			if 'vendor' == file:
				sg_dev_state.mSysFS_vendor, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, device_root + file )
			if 'type' == file:
				str_type, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, device_root + file )
				if str_type:
					sg_dev_state.mSysFS_type = int(str_type)

	return rc

####################################################################################################

SG_DEV_MISMATCH_NONE		= 0
SG_DEV_MISMATCH_VENDOR		= 1
SG_DEV_MISMATCH_MODEL		= 2
SG_DEV_MISMATCH_REV			= 3
SG_DEV_MISMATCH_TYPE		= 4
SG_DEV_MISMATCH_LU_CONN		= 5
SG_DEV_MISMATCH_LU_DISCONN	= 6
SG_DEV_MISMATCH_SIZE		= 7
SG_DEV_MISMATCH_NODATA		= 100

####################################################################################################
def isSysFS_stale(vi_errorAndLog_obj, sg_dev_state):
	"""
	Determines if the SysFs entries are stale and not up-to-date as what the SCSI INQ commands
	returned.

	Return value can be:
	SG_DEV_MISMATCH_NONE = 0
	SG_DEV_MISMATCH_VENDOR = 1
	SG_DEV_MISMATCH_MODEL = 2
	SG_DEV_MISMATCH_REV = 3
	SG_DEV_MISMATCH_TYPE = 4
	SG_DEV_MISMATCH_LU_CONN = 5
	SG_DEV_MISMATCH_LU_DISCONN = 6
	SG_DEV_MISMATCH_SIZE = 7
	SG_DEV_MISMATCH_NODATA = 100

	"""
#
	if not sg_dev_state:
		return SG_DEV_MISMATCH_NODATA

	if sg_dev_state.mSysFS_vendor != sg_dev_state.mSCSI_vendor:
		vi_errorAndLog_obj.vi_print_dbg(2, "Vendor mismatch: SysFS:%s != SCSI:%s\n" % (sg_dev_state.mSysFS_vendor, sg_dev_state.mSCSI_vendor))
		return SG_DEV_MISMATCH_VENDOR
	# This weeds out the EMC arrays. The model changes from 'LUNZ' to 'RAID' when the LUN
	# is attached (The old 'RAID' is of type disk..)
	if sg_dev_state.mSysFS_model != sg_dev_state.mSCSI_model:
		vi_errorAndLog_obj.vi_print_dbg(2, "Model mismatch: SysFS:%s != SCSI:%s\n" % (sg_dev_state.mSysFS_model, sg_dev_state.mSCSI_model))
		return SG_DEV_MISMATCH_MODEL
	if sg_dev_state.mSysFS_rev != sg_dev_state.mSCSI_rev:
		vi_errorAndLog_obj.vi_print_dbg(2, "Rev mismatch: SysFS:%s != SCSI:%s\n" % (sg_dev_state.mSysFS_rev, sg_dev_state.mSCSI_rev))
		return SG_DEV_MISMATCH_REV
	# This weeds out NETAPP arrays. The device type alters from 'no device type' (1Fh) to 'disk' (00h)
	# when the LUN is recycled.
	if sg_dev_state.mSysFS_type != sg_dev_state.mSCSI_type:
		vi_errorAndLog_obj.vi_print_dbg(2, "SCSI type mismatch: SysFS:%d != SCSI:%d\n" % (sg_dev_state.mSysFS_type, sg_dev_state.mSCSI_type))
		return SG_DEV_MISMATCH_TYPE
	# CD-ROMs size is 0 and non-connected LUNs are -1 (sg_readcap failed)
	if sg_dev_state.mSysFS_size > 0 and sg_dev_state.mSCSI_size > 0:
		if sg_dev_state.mSysFS_size != sg_dev_state.mSCSI_size:
			vi_errorAndLog_obj.vi_print_dbg(2, "Size mismatch: SysFS:%s != SCSI:%s\n" % (sg_dev_state.mSysFS_size, sg_dev_state.mSCSI_size))
			return SG_DEV_MISMATCH_SIZE
	# This weeds out Compellant arrays changing state. We use the Peripheral Qualifier to figure
	# out if this device is connected. If we expect it to be connected (b/c it has a block device associated) but
	# find out that the LU is disconnected, or vice-versa we need to refresh. Look in SPC3, section 6.4.1.
	# Note: PQual is 0 if the LU is connected.
	if sg_dev_state.mBlkName:
		if sg_dev_state.mPQual != 0:
			vi_errorAndLog_obj.vi_print_dbg(2, "Device %s is expected to be connected to the logical unit (%s), but it is disconnected!\n" \
				 % (sg_dev_state.mBlkName, sg_dev_state.mSCSI_device_id.as_str()))
			return SG_DEV_MISMATCH_LU_CONN
	else:
		if sg_dev_state.mPQual == 0:
			vi_errorAndLog_obj.vi_print_dbg(2, "Device %s is expected to be disconnected to the logical unit (%s), but it is connected!\n" \
				 % (sg_dev_state.mDeviceName, sg_dev_state.mSCSI_device_id.as_str()))
			return SG_DEV_MISMATCH_LU_DISCONN

	vi_errorAndLog_obj.vi_print_dbg(2, "SysFS has up-to-date information regarding %s" % (sg_dev_state.mDeviceName))
	return SG_DEV_MISMATCH_NONE

####################################################################################################
def sg_device_disconnect(vi_errorAndLog_obj, sg_dev_state, forceDisconnection ):
	"""
	Disconnects a SCSI generic device using the SysFS.
	"""
#
	sg_delete_path = "/sys/class/scsi_device/%s/device/delete" % ( sg_dev_state.mSCSI_device_id.as_str())
	if not os.path.exists ( sg_delete_path ):
		vi_errorAndLog_obj.vi_print_dbg(2, "Device %s doesn't exist? (%s)" % (sg_dev_state.mDeviceName, sg_delete_path));
		return 0

	sg_device = "/dev/%s" % (sg_dev_state.mDeviceName)
	# Flush out the dirty blocks. This is imperative otherwise we would have I/O in the dirty cache
	# which hasn't yet reached the SCSI ml.
	st_utils.vi_exe_sync( vi_errorAndLog_obj, 'sync' )

	# The deletion will flush out any outstanding blocks that have hit the SCSI middlelayer. It will
	# also invoke an SCSI SYNC command.
	write_failed	= st_utils.vi_write_text_value( vi_errorAndLog_obj, sg_delete_path, '1', False )

	st_utils.vi_exe_sync( vi_errorAndLog_obj, 'sync' )

	# This is the kicker.. /dev/<something> doesn't immediately disappear. Try up to 10 seconds
	mRetries = 10
	while (mRetries > 0):
		if not os.path.exists( sg_device ):
			break
		time.sleep ( 1 )
		mRetries -= 1

	if os.path.exists( sg_device ):
		vi_errorAndLog_obj.vi_print_dbg( 3, "SCSI generic device disconnect failed : %s" % ( sg_device ) )
		return 1

	return 0

####################################################################################################
def sg_device_rescan(vi_errorAndLog_obj, sg_dev_state, forceDisconnection ):
	"""
	Rescans a SCSI generic device using the SysFS.
	"""
#
	sg_rescan_path = "/sys/class/scsi_generic/%s/device/rescan" % ( sg_dev_state.mDeviceName)
	if not os.path.exists ( sg_rescan_path ):
		vi_errorAndLog_obj.vi_print_dbg(2, "Device %s doesn't exist? (%s)" % (sg_dev_state.mDeviceName, sg_rescan_path));
		return 1

	write_failed	= st_utils.vi_write_text_value( vi_errorAndLog_obj, sg_rescan_path, '1', False )

	if write_failed:
		vi_errorAndLog_obj.vi_print_dbg( 3, "SCSI generic device rescan failed : %s" % ( sg_device ) )
		return 1

	return 0

####################################################################################################
class SG_device_Operations(BlkDev_Operations):

	"""SG device class for BlkDev_Operations
	"""

	########################################
	def __init__( self, blkDevInfo, sg_device = None ):

		self.mBlkDevInfo	= blkDevInfo
		if sg_device:
			self.mSG_device		= sg_device # Which points to the SG_device_State_OBJ
		else:
			self.mSG_device		= SG_device_State_OBJ()

	########################################
	def __repr__( self ):

		return "SG_device__Operations[%s]" % ( self.mBlkDevInfo.mBlkDev )

	########################################
	def disconnect_block_device( self, vi_errorAndLog_obj, forceDisconnection ):
		"""
		Disconnects a SCSI generic device using the SysFS.
		"""
		return sg_device_disconnect(vi_errorAndLog_obj, self.mSG_device, forceDisconnection)

	########################################
	def revalidate_block_device( self, vi_errorAndLog_obj ):
		"""
		Updates a SCSI generic device to the latest. Returns zero if succesfull. Otherwise
		an errno number.

		"""

		rc = errno.ENODEV
		if not self.mSG_device.mSCSI_device_id:
			# Figure out that from mBlkDevInfo.mSCSIdevID
			if self.mBlkDevInfo.mSCSIdevID:
				self.mSG_device.mSCSI_device_id = self.mBlkDevInfo.mSCSIdevID
			else:
				return rc

		rc = do_sysfs_inq ( vi_errorAndLog_obj, self.mSG_device)
		if rc:
			return rc

		rc = do_scsi_inq ( vi_errorAndLog_obj, self.mSG_device)

		return rc

	########################################
	def	potentiallySAN( self ):
		"""
		Finds out if the SG device has MultiP or TGPS bit set and if so returns True.
		Not all SAN disks (like MSA1000) have these bits set. Some of the iSCSI enclosures don't
		set it unless you configure the iSCSI disk to be shared between more than one IQN (DS3300)
		"""

		if self.mSG_device.mDeviceName:
			if self.mSG_device.mMultiP>0 or self.mSG_device.mTGPS>0 or self.mSG_device.mRDAC>0:
				return True
		return False
	########################################
	def is_stale ( self, vi_errorAndLog_obj):
		"""
		Finds out if the SysFS entries are stale compared to the SCSI INQ results.
		"""
		if self.mSG_device.mDeviceName:
			return isSysFS_stale(vi_errorAndLog_obj, self.mSG_device) != 0
		return False

####################################################################################################

####################################################################################################
def find_sg_devices_on_host( vi_errorAndLog_obj, scsi_host):
	"""
	This routine finds the SCSI Generic devices on a particular host. Returns a list of
	SG_device_State_OBJ objects.
	"""
#
	sg_devices_LIST = []

	scsi_dev_root = "/sys/class/scsi_device/"
	for scsi_target in os.listdir( scsi_dev_root ):
		sg_dev_state = 	SG_device_State_OBJ()
		sg_dev_state.mSCSI_device_id = gen_SCSI_device_ID_OBJ_from_str(scsi_target)
		if sg_dev_state.mSCSI_device_id and sg_dev_state.mSCSI_device_id.mHost == int(scsi_host):
			# And the real fun starts here.
			if do_sysfs_inq ( vi_errorAndLog_obj, sg_dev_state) == 0:
				if do_scsi_inq ( vi_errorAndLog_obj, sg_dev_state) == 0:
					vi_errorAndLog_obj.vi_print_dbg(4, "Newly discovered: %s\n" % (repr(sg_dev_state)))
					sg_devices_LIST.append (sg_dev_state)

	# Update the dictionary.
	return sg_devices_LIST

####################################################################################################
def disconnect_stale_sg_devices_on_host( vi_errorAndLog_obj, SCSIHostNum):
	"""
	Disconnects SCSI generic (and its block if they exist) devices if the SysFS information is
	out of date compared to the information retrieved via the SCSI commands.

	If the block devices are under multipath control, then rescan the block device if SysFS
	deviates from SCSI commands info (right now just the size).

	Returns:
	Number of SCSI devices that were disconnected and number of SCSI devices rescanned.
	"""
#
	sg_dev_LIST = find_sg_devices_on_host(vi_errorAndLog_obj, SCSIHostNum)
	sg_count = 0
	rescan_count = 0
	dm_blk_dev_DICT = get_blk_dev_to_dm_map_lookup_dict(vi_errorAndLog_obj, True)
	for sg_dev_state in sg_dev_LIST:
		dm_blk_dev = None
		if isSysFS_stale(vi_errorAndLog_obj, sg_dev_state):
			if sg_dev_state.mBlkName:
				# Need to check device mapper to see if this is in usage.
				dm_blk_dev = dm_blk_dev_DICT.get(sg_dev_state.mBlkName, None)

			vi_errorAndLog_obj.vi_print_log( "Kernel out of sync for (%s) Re-syncing" % ( str(sg_dev_state) ))

			# If inside a multipath device, rescanning is fine.
			if dm_blk_dev:
				if sg_device_rescan(vi_errorAndLog_obj, sg_dev_state, False) != 0:
					vi_errorAndLog_obj.vi_print_wrn( "Failed to rescan the kernel for (%s)" % ( str(sg_dev_state) ))
				else:
					rescan_count += 1

				# iregardless, we don't want to disconnect the block device.
				continue

			if sg_device_disconnect(vi_errorAndLog_obj, sg_dev_state, False) != 0:
				vi_errorAndLog_obj.vi_print_wrn( "Failed to re-sync the kernel for (%s)" % ( str(sg_dev_state) ))
			else:
				sg_count += 1

	return sg_count, rescan_count

####################################################################################################
def find_sg_devices( vi_errorAndLog_obj):
	"""
	This routine finds all of the SCSI Generic devices. Returns a dictionary of all
	SG_device_State_OBJ wherin in the key is the SCSI host number as so:
		0 -> [mDeviceName: sg0 mBlkName: sr0 mPQual: 0 mTGPS: 0 mRDAC: -1 mMultiP: 0 (SCSI: mVendor: LITE-ON mModel: CD-ROM LTN-4891S, mRev: NDS3, mType: 5,mSize: -1;SysFS: mVendor: LITE-ON, mModel: CD-ROM LTN-4891S, mRev: NDS3, mType: 5, mSize: 2097151) mSCSI_Dev_ID: 0:0:0:0: ]
	"""
#
	sgdev_DICT = {}

	scsi_dev_root = "/sys/class/scsi_device/"
	for scsi_target in os.listdir( scsi_dev_root ):
		sg_dev_state = 	SG_device_State_OBJ()
		sg_dev_state.mSCSI_device_id = gen_SCSI_device_ID_OBJ_from_str(scsi_target)
		if sg_dev_state.mSCSI_device_id:
			# And the real fun starts here.
			if do_sysfs_inq ( vi_errorAndLog_obj, sg_dev_state) == 0:
				if do_scsi_inq ( vi_errorAndLog_obj, sg_dev_state) == 0:
					vi_errorAndLog_obj.vi_print_dbg(4, "Newly discovered: %s\n" % (repr(sg_dev_state)))
					sg_devices_LIST = sgdev_DICT.get( sg_dev_state.mSCSI_device_id.mHost, None )
					if not sg_devices_LIST:
						sgdev_DICT [ sg_dev_state.mSCSI_device_id.mHost ] = [sg_dev_state]
					else:
						sg_devices_LIST.append (sg_dev_state)

	return sgdev_DICT

####################################################################################################
def get_sgdev_to_host_nr ( vi_errorAndLog_obj):
	"""
	Returns a dictionary wherein the key is the SCSI generic device name and the value is the
	host nunber as so:

		sg1 -> 0
		sg2 -> 0
		sg3 -> 4
		sg5 -> 10
	"""
#
	sg_devices_DICT = find_sg_devices ( vi_errorAndLog_obj)

	sg_dev_to_host_nr_DICT = {}

	for host_nr, sg_dev_LIST in sg_devices_DICT.iteritems():
		for sg_dev_state in sg_dev_LIST:
			sg_dev_to_host_nr_DICT[ sg_dev_state.mDeviceName ] = host_nr

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, sg_dev_to_host_nr_DICT )

	return sg_dev_to_host_nr_DICT

####################################################################################################
def get_host_nr_to_sgdev ( vi_errorAndLog_obj):
	"""
	Returns a dictionary wherein the key is the SCSI host number and the value is the
	SCSI generic name as so:

		0 -> (sg1,sg2)
		4 -> (sg3)
		10 -> (sg5)
	"""
#

	sg_devices_DICT = find_sg_devices ( vi_errorAndLog_obj )

	host_nr_to_sgdev_DICT = {}

	for host_nr, sg_dev_LIST in sg_devices_DICT.iteritems():
		for sg_dev_state in sg_dev_LIST:
			sg_devices_LIST = host_nr_to_sgdev_DICT.get(host_nr, None)
			if not sg_devices_LIST:
				host_nr_to_sgdev_DICT[ host_nr ] = [ sg_dev_state.mDeviceName ]
			else:
				sg_devices_LIST.append( sg_dev_state.mDeviceName )

	if 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, host_nr_to_sgdev_DICT )

	return host_nr_to_sgdev_DICT

####################################################################################################
# END
