#!/usr/bin/env python
#
# Copyright (C) 2011-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Konrad Rzeszutek
#
"""Utilities for dealing with loop device.
"""

import st_utils

import errno
import time
import re
import traceback
import os
####################################################################################################

__all__ = [	'LoopBack_BlkDev_OBJ'
		,	'LoopBack_BlkDev_Operations'
		,	'find_loopback_disks'
		,	'loopback_is_occupied'
		,	'loopback_mount']

####################################################################################################

####################################################################################################

class LoopBack_BlkDev_OBJ:
	"""
	Contains basic data about the loop back device.
	"""
	def __init__(self):
		self.mBlkDev		= None
		self.mDevPath		= ""
		self.mWWID			= ""
		self.mNumber		= -1
		self.mEndOffset		= -1 # Offsets are in 512K sizes
		self.mBegOffset		= -1
		self.mSource		= None
		self.mSizeInBytes	= -1
		self.mDevNums		= None # In major:minor string format.
	def __repr__(self):
		str = "/dev/loop%d (%s), size:%d" % (
						self.mNumber
					,	self.mDevNums
					,	self.mSizeInBytes)
		if self.mWWID:
			str = str + " (%s->%s)" % (self.mSource ,self.mWWID)
		else:
			str = str + " (%s)" % (self.mSource)

		if self.mBegOffset or self.mEndOffset:
			str=str+" offset(beg:%d,end:%d)" % (self.mBegOffset, self.mEndOffset)
		return str

####################################################################################################
class LoopBack_BlkDev_Operations:
	"""
	The loopback operations.
	"""

	def __init__(self, blkDevInfo, loopback_disk):
                if blkDevInfo.mTransport != 'loopback':
                        raise ValueError, "given blkDevInfo %s is not loopback [%s] !?!" % ( blkDevInfo.mBlkDev, blkDevInfo.mTransport )

                self.mBlkDevInfo        = blkDevInfo
                self.mLoopback           = loopback_disk

	def __repr__(self):
		return "LoopBack_BlkDev_Operations[%s]" % (self.mBlkDevInfo.str())

	def	disconnect_block_device( self, vi_errorAndLog_obj, forceDisconnection = False):
		"""
		Disconnect the loop-back device. Returns zero if successfull otherwise a positive digit.
		"""
		if self.mLoopback.mWWID:
			# VIKR: We should check if the WWID exists physically
			dmDeleteCmd = ['/sbin/dmsetup', 'remove', self.mLoopback.mWWID]
			stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync ( vi_errorAndLog_obj, dmDeleteCmd)
			if len (stdout_LIST) > 0 or len (stderr_LIST) > 0:
				vi_errorAndLog_obj.vi_print_err("Failed to execute: %s " % (dmDeleteCmd))
				vi_errorAndLog_obj.vi_print_err("".join(stdout_LIST)+"".join(stderr_LIST))
		if (loopback_is_occupied ( vi_errorAndLog_obj, self.mLoopback.mNumber ) ):
			loopStrCmd = "/sbin/losetup -d /dev/loop%d" % (self.mLoopback.mNumber)
			stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync ( vi_errorAndLog_obj, loopStrCmd )
			if not (loopback_is_occupied ( vi_errorAndLog_obj, self.mLoopback.mNumber ) ):
				return 0
		return 1

	def	revalidate_block_device ( self, vi_errorAndLog_obj ) :
		"""
		Updates the loop back device to the latests sysfs values.
		"""
		blkdev_info_path  = "/sys/block/loop%d" % ( self.mLoopback.mNumber)
		if not os.path.exists ( blkdev_info_path ):
			return 1

		blkdev_info_path = "/sys/block/loop%d/dev" % ( self.mLoopback.mNumber )
		self.mLoopback.mDevNums, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
		if read_failed:
			vi_errorAndLog_obj.vi_print_err( "blkdev 'loop%d' vanished [%s] !?!" % ( self.mLoopback.mNumber, blkdev_info_path ) )
			return 1

		blkdev_info_path = "/sys/block/%d/size" % ( self.mLoopback.mNumber )
		blkdev_sizeInSectors, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
		if read_failed:
			vi_errorAndLog_obj.vi_print_err( "blkdev 'loop%d' vanished [%s] !?!" % ( self.mLoopback.mNumber, blkdev_info_path ) )
			return 1

		self.mLoopback.mSizeInBytes	= int( blkdev_sizeInSectors ) * 512

		return 0

####################################################################################################

global__cached__loopback_disk_LIST	= None

def find_loopback_disks ( vi_errorAndLog_obj, refresh ):
	"""
	Find the existing loopback devices and return a list of LoopBack_BlkDev_OBJs
	"""

	global global__cached__loopback_disk_LIST
	if not refresh and global__cached__loopback_disk_LIST:
		return global__cached__loopback_disk_LIST

	loopback_disk_LIST = []
	try:
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync( vi_errorAndLog_obj, '/sbin/losetup -a', 20 )
		#/dev/loop0: [0011]:14139 (/tmp/primary_network_info) offset=12
		loopback_disk_cmpldRegEx = re.compile( '\/dev\/loop(\d+):.*\((.*)\)' )
		loopback_disk_offsetRegEx = re.compile( '.*offset\s(\d+)' )
		del( stderr_LIST )
		for loopback_line in stdout_LIST:
			loopback_disk_match 		= loopback_disk_cmpldRegEx.match ( loopback_line.rstrip() )
			if not loopback_disk_match:
				print vi_errorAndLog_obj.vi_print_err( "Couldn't parse correctly: \"%s\". Skipping." % (loopback_line.rstrip()))
				continue
			lb_number		= int(loopback_disk_match.group(1))
			lb_source		= loopback_disk_match.group(2)
			lb_beg_offset	= 0
			loopback_disk_match_offset 	= loopback_disk_offsetRegEx.match (loopback_line.rstrip() )
			if loopback_disk_match_offset:
				lb_beg_offset	= int(loopback_disk_match_offset.group(1))

			blkdev_info_path = "/sys/block/loop%d/size" % (lb_number)
			loopback_disk_sizeInSectors, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
			if read_failed:
				vi_errorAndLog_obj.vi_print_err( "blkdev 'loop%d' vanished [ %s] !?!" % (lb_number, blkdev_info_path))
				return None
			lb_sizeInBytes	= int(loopback_disk_sizeInSectors) * 512

			blkdev_info_path = "/sys/block/loop%d/dev" % ( lb_number )
			lb_devNums, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
			if read_failed:
				vi_errorAndLog_obj.vi_print_err( "blkdev 'loop%d' vanished [ %s] !?!" % (lb_number, blkdev_info_path))
				return None
			loopback_disk			= LoopBack_BlkDev_OBJ()
			loopback_disk.mBlkDev	= "loop%d" % (lb_number)
			loopback_disk.mDevPath	= "/dev/loop%d" % (lb_number)
			loopback_disk.mNumber	= lb_number
			loopback_disk.mSource	= lb_source
			loopback_disk.mBegOffset	= lb_beg_offset
			# We don't figure this out here yet. This requires parsing the device mapper table and see what
			# the size has been specified as and figure out the difference.
			loopback_disk.mEndOffset	= 0
			loopback_disk.mSizeInBytes = lb_sizeInBytes
			loopback_disk.mDevNums	= lb_devNums
			loopback_disk_LIST.append ( loopback_disk )
	except:
		vi_errorAndLog_obj.vi_print_err( "Encountered problems extracting loopback disk info\n%s" % (traceback.format_exc()))
		return []

	#vi_errorAndLog_obj.vi_print_dbg(0, loopback_disk_LIST )

	global__cached__loopback_disk_LIST = loopback_disk_LIST
	return global__cached__loopback_disk_LIST

def loopback_mount_disk ( vi_errorAndLog_obj, sourceFile, begOffset = 0):
	"""
	Mounts the passed in file/disk to a loopback device. The returned
	value is the newly mounted LoopBack_BlkDev_OBJ or None if didn't
	succeed. VIKR, erro codes?
	"""
	loopback_disk_LIST = find_loopback_disks( vi_errorAndLog_obj , False)
	number = 0
	if len(loopback_disk_LIST):
		number = loopback_disk_LIST[len(loopback_disk_LIST)-1].mNumber
		number += 1

	if number > 255 and number < 0:
		return None
		#return errno.EADDRNOTAVAIL

	if not os.path.exists ( sourceFile ):
		return None
		#return errno.ENOENT

	cmdStr = "/sbin/losetup /dev/loop%d %s" % (number, sourceFile)
	if begOffset:
		cmdStr += " -o %d" % (begOffset)
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync ( vi_errorAndLog_obj, cmdStr )
	if exitStatus:
		return None
		#return exitStatus


	blkdev_info_path = "/sys/block/loop%d/size" % (number)
	loopback_disk_sizeInSectors, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
	if read_failed:
		vi_errorAndLog_obj.vi_print_err( "blkdev 'loop%d' vanished [ %s] !?!" % (lb_number, blkdev_info_path))
		return None
	lb_sizeInBytes	= int(loopback_disk_sizeInSectors) * 512

	blkdev_info_path = "/sys/block/loop%d/dev" % ( number )
	lb_devNums, read_failed = st_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
	if read_failed:
		vi_errorAndLog_obj.vi_print_err( "blkdev 'loop%d' vanished [ %s] !?!" % (lb_number, blkdev_info_path))
		return None

	blkDev = LoopBack_BlkDev_OBJ()
	blkDev.mSource = sourceFile
	blkDev.mBegOffset = begOffset
	blkDev.mNumber = number
	blkDev.mDevNums = lb_devNums
	blkDev.mSizeInBytes = lb_sizeInBytes

	return blkDev

def loopback_is_occupied ( vi_errorAndLog_obj, loopBackNumber ):
	"""
	Finds out if the loopback device is occupied.
	"""
	loopStrCmd = "/sbin/losetup /dev/loop%d" % (loopBackNumber)
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync ( vi_errorAndLog_obj, loopStrCmd)

	return (exitStatus == 0)

def loopback_create_map ( vi_errorAndLog_obj, blkDev , name, endOffset = 0):
	"""
	Create a device mapper map for the loopback device. Returns zero
	if succesfull.

	This routine invokes this application:
	dmsetup create <name> --table "0 XXX linear /dev/loopY 0"

	where:
		<name>	is the 'name' argument passed in.
		XXXX	is the size of the device. Extracted from blkDev
		Y		is the loopback device number. Extracted from blkDev

	"""

	if blkDev.mSizeInBytes > 0 and blkDev.mNumber >= 0:
		table = "0 %d linear /dev/loop%d 0" % (int(blkDev.mSizeInBytes / 512) - endOffset, blkDev.mNumber)
		dmSetupCmd = ['/sbin/dmsetup', 'create',name, '--table', table]
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = st_utils.vi_exe_sync ( vi_errorAndLog_obj, dmSetupCmd)
		if len (stdout_LIST) > 0 or len (stderr_LIST) > 0:
			vi_errorAndLog_obj.vi_print_err("Failed to execute: %s " % (dmSetupCmd))
			vi_errorAndLog_obj.vi_print_err("".join(stdout_LIST)+"".join(stderr_LIST))
			return -1
		blkDev.mWWID = name
		blkDev.mEndOffset = endOffset
		return exitStatus;
	return 1

def loopback_mount ( vi_errorAndLog_obj, source, name, begOffset = 0, endOffset = 0):
	"""
	Mount the source file as a loopback device and create a device mapper map for it.
	Returns the LoopBack_BlkDev_OBJ.
	"""

	if begOffset < 0 or endOffset < 0:
		vi_errorAndLog_obj.vi_print_err("loopback_mount (%s, %s, %d, %d). Offset values are wrong." % (source, name, begOffset, endOffset))
		return None
	blkDev =  loopback_mount_disk ( vi_errorAndLog_obj, source, begOffset )
	if (blkDev):
		if not (loopback_create_map ( vi_errorAndLog_obj, blkDev, name, endOffset )):
			return blkDev

	return None

# END
####################################################################################################
