#!/usr/bin/env python
#
# Copyright (C) 2011-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas, Konrad Rzeszutek
#
"""Utilities for dealing with udev
"""

import st_utils

import time
import errno

####################################################################################################

__all__ = [	'udev_idle' ]

####################################################################################################
def udev_idle( vi_errorAndLog_obj, idleTimeSpanInSecs, timeoutInSecs ):
	"""
	This routine with wait for udev events to be idle for at least idleTimeSpanInSecs
	and will timeout when timeoutInSecs have elapsed.
	"""
#
	udevSeqNumStr, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, '/sys/kernel/uevent_seqnum')
	if read_failed:
		vi_errorAndLog_obj.vi_print_err( "read 1 of '/sys/kernel/uevent_seqnum' failed!" )
		return errno.ETIMEDOUT

	startTimeInSecs		= time.time()
	startingSeqNum		= int( udevSeqNumStr )
	spanStartTimeInSecs	= startTimeInSecs
	spanStartSeqNum		= startingSeqNum

	exitStatus			= 0
	while True:

		nowTimeInSecs	= time.time()
		udevSeqNumStr, read_failed = st_utils.vi_read_text_value(vi_errorAndLog_obj, '/sys/kernel/uevent_seqnum')
		if read_failed:
			vi_errorAndLog_obj.vi_print_err( "read 2 of '/sys/kernel/uevent_seqnum' failed!" )
			return errno.ETIMEDOUT

		deltaTotalTimeInSecs	= st_utils.vi_calc_delta_time( startTimeInSecs, nowTimeInSecs )

		curSeqNum		= int( udevSeqNumStr )
		if curSeqNum >= startingSeqNum:
			deltaSeqNum	= curSeqNum - startingSeqNum
		else:
			deltaSeqNum	= startingSeqNum - curSeqNum

		# First see if udev has been idle
		if curSeqNum == spanStartSeqNum:

			deltaSpanTimeInSecs	= st_utils.vi_calc_delta_time( spanStartTimeInSecs, nowTimeInSecs )

			if deltaSpanTimeInSecs >= idleTimeSpanInSecs:

				vi_errorAndLog_obj.vi_print_log( "UDEVidle : idle %0.2f [ %0.2f : %s ] %s:%s" % \
										(	deltaSpanTimeInSecs
										,	deltaTotalTimeInSecs
										,	deltaSeqNum
										,	idleTimeSpanInSecs
										,	timeoutInSecs
										) )
				return 0	# success

		else:
			spanStartTimeInSecs	= nowTimeInSecs
			spanStartSeqNum		= curSeqNum

		# Check for timeout
		if deltaTotalTimeInSecs > timeoutInSecs:

			vi_errorAndLog_obj.vi_print_log( "UDEVidle : TIMED OUT (%s) [ %0.2f : %s ] %s:%s" % \
									(	exitStatus
									,	deltaTotalTimeInSecs
									,	deltaSeqNum
									,	idleTimeSpanInSecs
									,	timeoutInSecs
									) )
			break

		# Watch the udev event queue, and exits if all current events are handled
		cmd_LIST	= [ "/sbin/udevsettle", "--timeout=%s" % ( idleTimeSpanInSecs + 1 ) ]
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = \
 			st_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, idleTimeSpanInSecs + 4 )

	return errno.ETIMEDOUT

####################################################################################################
# END
