#!/usr/bin/env python
#
# Copyright (C) 2011-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for helping with reading and writing
"""

import traceback

from st_errorAndLog		import *

####################################################################################################

__all__ = [	'vi_read_text_value'
		,	'vi_read_text_lines'
		,	'vi_write_text_value'
		,	'vi_write_text_lines'
		,	'vi_read_write_text_lines' ]

####################################################################################################
def vi_read_text_value( vi_errorAndLog_obj, text_file_path ):
	"""
	This routine reads all of the lines of the text file at the given text_file_path and returns
	them as a single buffer

	Arguments:	vi_errorAndLog_obj	object to route errors to
				text_file_path		path to the text file to read

	NOTE:		lines are head tail whitespace stripped then concatenated with a single space

	Return:		read_buffer, did_read_fail_boolean
	"""
#
	final_text_value = ''
	try:
		text_FILE = open( text_file_path , 'r' )
		final_text_value = ' '.join( map( str.strip, text_FILE.readlines() ) )
		text_FILE.close()
	except IOError:
		vi_errorAndLog_obj.vi_print_dbg( 7, "open( %s ) failed" % ( text_file_path ) )
		return '', True
	except:
		vi_errorAndLog_obj.vi_print_err( "extraction from %s failed\n%s" % ( text_file_path, traceback.format_exc() ) )
		return '', True

	return final_text_value, False

####################################################################################################
def vi_read_text_lines( vi_errorAndLog_obj, text_file_path ):
	"""
	This routine reads all of the lines of the text file at the given text_file_path and returns
	them as a list of lines

	Arguments:	vi_errorAndLog_obj	object to route errors to
				text_file_path		path to the text file to read

	NOTE:		lines are head tail whitespace stripped

	Return:		readLines_LIST, did_read_fail_boolean
	"""
#
	text_lines_LIST = []
	try:
		text_FILE = open( text_file_path , 'r' )
		text_lines_LIST = map( str.strip, text_FILE.readlines() )
		text_FILE.close()
	except IOError:
		vi_errorAndLog_obj.vi_print_dbg( 7, "open( %s ) failed" % ( text_file_path ) )
		return [], True
	except:
		vi_errorAndLog_obj.vi_print_err( "extraction from %s failed\n%s" % ( text_file_path, traceback.format_exc() ) )
		return [], True

	return text_lines_LIST, False

################################################################################
def vi_write_text_value( vi_errorAndLog_obj, text_file_path, buffer_to_write, appendVStruncate ):
	"""
	This routine writes the given buffer to the text file specified

	Arguments:	vi_errorAndLog_obj	object to route errors to
				text_file_path		path to the text file to write
				buffer_to_write		data to write
				appendVStruncate	boolean whether to append or to truncate

	Return:		did_write_fail_boolean
	"""
#
	if appendVStruncate:
		open_mode = 'a'
	else:
		open_mode = 'w'

	try:
		text_FILE = open( text_file_path , open_mode )
		text_FILE.write( buffer_to_write )
		text_FILE.close()
	except IOError:
		vi_errorAndLog_obj.vi_print_dbg( 7, "open( %s ) failed" % ( text_file_path ) )
		return True
	except:
		vi_errorAndLog_obj.vi_print_err( "write to %s failed\n%s" % ( text_file_path, traceback.format_exc() ) )
		return True

	return False

################################################################################
def vi_write_text_lines( vi_errorAndLog_obj, text_file_path, text_lines_LIST, appendNewlineToEachLine, appendVStruncate ):
	"""
	This routine writes the given text lines to the text file specified

	Arguments:	vi_errorAndLog_obj		object to route errors to
				text_file_path			path to the text file to write
				text_lines_LIST			data to write
				appendNewlineToEachLine	append a '\n' to the end of each line
				appendVStruncate		boolean whether to append or to truncate

	Return:		did_write_fail_boolean
	"""
#
	if appendVStruncate:
		open_mode = 'a'
	else:
		open_mode = 'w'

	try:
		text_FILE = open( text_file_path , open_mode )
		if appendNewlineToEachLine:
			for line in text_lines_LIST:
				print >> text_FILE, line
		else:
			for line in text_lines_LIST:
				text_FILE.write( line )
		text_FILE.close()
	except IOError:
		vi_errorAndLog_obj.vi_print_dbg( 7, "open( %s ) failed" % ( text_file_path ) )
		return True
	except:
		vi_errorAndLog_obj.vi_print_err( "write to %s failed\n%s" % ( text_file_path, traceback.format_exc() ) )
		return True

	return False

################################################################################
def vi_read_write_text_lines( vi_errorAndLog_obj, src_text_file_path, trg_text_file_path, appendVStruncate, do_lstrip, do_rstrip ):
	"""
	This routine reads from the src and writes to the trg applying the given conversions.

	Arguments:	vi_errorAndLog_obj		object to route errors to
				src_text_file_path		path to the text file to read
				trg_text_file_path		path to the text file to write
				appendVStruncate		boolean whether to append or to truncate
				do_lstrip				boolean whether to lstrip each line
				do_rstrip				boolean whether to rstrip each line

	Return:		did_write_fail_boolean
	"""
#
	if appendVStruncate:
		open_mode = 'a'
	else:
		open_mode = 'w'

	read_LIST	= None
	try:
		read_FILE	= open( src_text_file_path, 'r' )
		read_LIST	= read_FILE.readlines()
		read_FILE.close()
	except IOError:
		vi_errorAndLog_obj.vi_print_dbg( 7, "open( %s ) failed" % ( src_text_file_path ) )
		return True
	except:
		vi_errorAndLog_obj.vi_print_err( "read from %s failed\n%s" % ( src_text_file_path, traceback.format_exc() ) )
		return True

	try:
		write_FILE	= open( trg_text_file_path, open_mode )
		if do_rstrip and not do_lstrip:
			for line in read_LIST:
				print >> write_FILE, line.rstrip()
		elif do_rstrip and do_lstrip:
			for line in read_LIST:
				print >> write_FILE, line.strip()
		elif not do_rstrip and do_lstrip:
			for line in read_LIST:
				print >> write_FILE, line.lstrip()
		else:
			for line in read_LIST:
				write_FILE.write( line )
		write_FILE.close()
		del( read_LIST )
	except IOError:
		del( read_LIST )
		vi_errorAndLog_obj.vi_print_dbg( 7, "open( %s ) failed" % ( trg_text_file_path ) )
		return True
	except:
		del( read_LIST )
		vi_errorAndLog_obj.vi_print_err( "write to %s failed\n%s" % ( trg_text_file_path, traceback.format_exc() ) )
		return True

	return False

####################################################################################################
# END
