#!/usr/bin/env python
#
# Copyright (C) 2011-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for packing and unpacking binary data
"""

import struct

####################################################################################################

__all__ = [	'vi_strctFrmt__bytes'
		,	'vi_strctFrmt__le__u8'
		,	'vi_strctFrmt__le__s8'
		,	'vi_strctFrmt__be__u8'
		,	'vi_strctFrmt__be__s8'
		,	'vi_strctFrmt__le__u16'
		,	'vi_strctFrmt__le__s16'
		,	'vi_strctFrmt__be__u16'
		,	'vi_strctFrmt__be__s16'
		,	'vi_strctFrmt__le__u32'
		,	'vi_strctFrmt__le__s32'
		,	'vi_strctFrmt__be__u32'
		,	'vi_strctFrmt__be__s32'
		,	'vi_strctFrmt__le__u64'
		,	'vi_strctFrmt__le__s64'
		,	'vi_strctFrmt__be__u64'
		,	'vi_strctFrmt__be__s64'
		,	'vi_structLayout' ]

####################################################################################################

class vi_strctFrmt__base:
	def GetDefaultValue( self ):
		return 0

####################################################################################################

class vi_strctFrmt__bytes(vi_strctFrmt__base):
	def __init__( self, count ):
		self.mCount = count
	def __repr__( self ):
		return "%ds" % ( self.mCount )
	def GetDefaultValue( self ):
		return chr(0) * self.mCount
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__le__u8(vi_strctFrmt__base):	# little endian :: unsigned 8-bit integer
	def __repr__( self ):
		return 'B'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__le__s8(vi_strctFrmt__base):	# little endian ::   signed 8-bit integer
	def __repr__( self ):
		return 'b'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__be__u8(vi_strctFrmt__base):	#   big endian :: unsigned 8-bit integer
	def __repr__( self ):
		return 'B'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>B', struct.pack( '@B', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@B', struct.pack( '>B', givenValue ) )[ 0 ]

class vi_strctFrmt__be__s8(vi_strctFrmt__base):	#   big endian ::   signed 8-bit integer
	def __repr__( self ):
		return 'b'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>b', struct.pack( '@b', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@b', struct.pack( '>b', givenValue ) )[ 0 ]

class vi_strctFrmt__le__u16(vi_strctFrmt__base):	# little endian :: unsigned 16-bit integer
	def __repr__( self ):
		return 'H'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__le__s16(vi_strctFrmt__base):	# little endian ::   signed 16-bit integer
	def __repr__( self ):
		return 'h'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__be__u16(vi_strctFrmt__base):	#   big endian :: unsigned 16-bit integer
	def __repr__( self ):
		return 'I'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>H', struct.pack( '@H', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@H', struct.pack( '>H', givenValue ) )[ 0 ]

class vi_strctFrmt__be__s16(vi_strctFrmt__base):	#   big endian ::   signed 16-bit integer
	def __repr__( self ):
		return 'i'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>h', struct.pack( '@h', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@h', struct.pack( '>h', givenValue ) )[ 0 ]

class vi_strctFrmt__le__u32(vi_strctFrmt__base):	# little endian :: unsigned 32-bit integer
	def __repr__( self ):
		return 'I'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__le__s32(vi_strctFrmt__base):	# little endian ::   signed 32-bit integer
	def __repr__( self ):
		return 'i'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__be__u32(vi_strctFrmt__base):	#   big endian :: unsigned 32-bit integer
	def __repr__( self ):
		return 'I'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>I', struct.pack( '@I', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@I', struct.pack( '>I', givenValue ) )[ 0 ]

class vi_strctFrmt__be__s32(vi_strctFrmt__base):	#   big endian ::   signed 32-bit integer
	def __repr__( self ):
		return 'i'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>i', struct.pack( '@i', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@i', struct.pack( '>i', givenValue ) )[ 0 ]

class vi_strctFrmt__le__u64(vi_strctFrmt__base):	# little endian :: unsigned 64-bit integer
	def __repr__( self ):
		return 'Q'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__le__s64(vi_strctFrmt__base):	# little endian ::   signed 64-bit integer
	def __repr__( self ):
		return 'q'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return givenValueAsNative
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return givenValue

class vi_strctFrmt__be__u64(vi_strctFrmt__base):	#   big endian :: unsigned 64-bit integer
	def __repr__( self ):
		return 'Q'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>Q', struct.pack( '@Q', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@Q', struct.pack( '>Q', givenValue ) )[ 0 ]

class vi_strctFrmt__be__s64(vi_strctFrmt__base):	#   big endian ::   signed 64-bit integer
	def __repr__( self ):
		return 'q'
	def ConvertFromNativeEndianLayout( self, givenValueAsNative ):
		return struct.unpack( '>q', struct.pack( '@q', givenValueAsNative ) )[ 0 ]
	def ConvertIntoNativeEndianLayout( self, givenValue ):
		return struct.unpack( '@q', struct.pack( '>q', givenValue ) )[ 0 ]

####################################################################################################

class vi_structValue_OBJ:

	def __init__( self, packTPLidx, theValue, visibleIdx, strctFrmtSubClass ):

		self.mPackTPLidx			= packTPLidx
		self.mTheValue				= theValue
		self.mVisibleIdx			= visibleIdx
		self.mStrctFrmtSubClass		= strctFrmtSubClass

	def __repr__( self ):

		return "pIdx: %s V: %s vIdx: %s frmt: %s" % \
					(	self.mPackTPLidx
					,	self.mTheValue
					,	self.mVisibleIdx
					,	self.mStrctFrmtSubClass )

####################################################################################################

class vi_structLayout:

	"""Class assists with managing binary struct layouts and packing into and out of them

	Arguments:	frmtFieldname_TPL_LIST	LIST of ( format, fieldName, optionalIsHiddenBool ) tuples
				requiredByteLen			number of bytes that the given frmtFieldname_TPL_LIST
										MUST be equal to ( else caller bug )

	Example:

	example_hdr_frmtFieldname_TPL_LIST = [
		(	vi_strctFrmt__be__u32(),	'magic_1'		,	True	)	# offset ==   0 |  4 bytes :: Magic 1 ( 0xBAADF00D big endian )
	,	(	vi_strctFrmt__le__s32(),	'verNum'					)	# offset ==   4 |  4 bytes :: Major Version Num
	,	(	vi_strctFrmt__le__u64(),	'dataOffset'				)	# offset ==   8 |  8 bytes :: Major Version Num
	,	(	vi_strctFrmt__le__s64(),	'bigValue'					)	# offset ==  16 |  8 bytes :: Major Version Num
	,	(	vi_strctFrmt__bytes( 40 ),	'reserved'		,	True	)	# offset ==  24 | 40 bytes :: reserved ( zero )
	]																	# Total ==>  64 bytes ( fixed size hdr )

	example_hdr_layout = vi_structLayout( example_hdr_frmtFieldname_TPL_LIST, 64 )

	allSetBuffer = chr(0xFF) * 64

	example_hdr_layout.IngestBuffer( allSetBuffer )
	"""

	def __init__( self, frmtFieldname_TPL_LIST, requiredByteLen ):

		structFrmt	= '@'	# Always ingest/export as native endian, and convert from/into there
		for frmtFieldname_TPL in frmtFieldname_TPL_LIST:
			if not isinstance( frmtFieldname_TPL[1], str ):
				raise ValueError, "given frmtFieldname_TPL[1] MUST be a string"
			if not isinstance( frmtFieldname_TPL[0], vi_strctFrmt__base ):
				raise ValueError, "given '%s' frmtFieldname_TPL[0] '%s' MUST be a vi_strctFrmt__..." % \
									( frmtFieldname_TPL[1], type( frmtFieldname_TPL[0] ) )

			structFrmt	= structFrmt + str(frmtFieldname_TPL[0])

		self.mStructSizeInBytes	= struct.calcsize( structFrmt )
		if self.mStructSizeInBytes != requiredByteLen:
			raise ValueError, "given frmt_name_TPL_LIST size %d != %d requiredByteLen" % \
								( self.mStructSizeInBytes, requiredByteLen )

		self.mStruct		= structFrmt	# Struct( structFrmt ) Magnesium is only python 2.4
		self.mValues_DICT	= {}
		idx					= 0
		visIdx				= 0
		for frmtFieldname_TPL in frmtFieldname_TPL_LIST:
			hidden	= False
			if len(frmtFieldname_TPL) < 3 or not frmtFieldname_TPL[2]:
				structValue	= vi_structValue_OBJ( idx, frmtFieldname_TPL[0].GetDefaultValue(), visIdx, frmtFieldname_TPL[0] )
				visIdx		= visIdx + 1
			else:
				structValue = vi_structValue_OBJ( idx, frmtFieldname_TPL[0].GetDefaultValue(),   None, frmtFieldname_TPL[0] )
			self.mValues_DICT[ frmtFieldname_TPL[1] ]	= structValue
			idx											= idx + 1
		self.mNumVisible	= visIdx

	########################################
	def Get( self, attrname ):
		structValue	= self.mValues_DICT.get( attrname, None )
		if not structValue:
			raise AttributeError, "Give attrname '%s' does not exist" % ( attrname )
			return None
		return structValue.mTheValue

	########################################
	def Set( self, attrname, newValue ):
		structValue	= self.mValues_DICT.get( attrname, None )
		if not structValue:
			raise AttributeError, "Give attrname '%s' does not exist" % ( attrname )
			return True
		structValue.mTheValue	= newValue
		return False

	########################################
	def Clear( self ):
		for fieldName, structValue in self.mValues_DICT.iteritems():
			structValue.mTheValue	= structValue.mStrctFrmtSubClass.GetDefaultValue()

	########################################
	def IngestBuffer( self, givenSrcBuffer, srcOffset=0 ):
#		newValues_TPL = self.mStruct.unpack_from( givenSrcBuffer, srcOffset ) # Magnesium is only python 2.4
		if srcOffset != 0:
			raise ValueError, "srcOffset MUST be 0 until Sodium"
		newValues_TPL = struct.unpack( self.mStruct, givenSrcBuffer )
		for fieldName, structValue in self.mValues_DICT.iteritems():
			rawNewValue				= newValues_TPL[ structValue.mPackTPLidx ]
			convertednewValue		= structValue.mStrctFrmtSubClass.ConvertFromNativeEndianLayout( rawNewValue )
			structValue.mTheValue	= convertednewValue

	########################################
	def GetAsBuffer( self ):
		curValues_LIST	= range( len( self.mValues_DICT ) )
		for fieldName, structValue in self.mValues_DICT.iteritems():
			curValue	= structValue.mTheValue
			rawValue	= structValue.mStrctFrmtSubClass.ConvertIntoNativeEndianLayout( curValue )
			curValues_LIST[  structValue.mPackTPLidx ] = rawValue
		return struct.pack( self.mStruct, *curValues_LIST )

	########################################
	def FillBuffer( self, givenTrgBuffer, trgOffset=0 ):
		raise ValueError, "not available until Sodium"
		curValues_LIST	= range( len( self.mValues_DICT ) )
		for fieldName, structValue in self.mValues_DICT.iteritems():
			curValue	= structValue.mTheValue
			rawValue	= structValue.mStrctFrmtSubClass.ConvertIntoNativeEndianLayout( curValue )
			curValues_LIST[  structValue.mPackTPLidx ] = rawValue
		struct.pack( self.mStruct, *curValues_LIST )
		self.mStruct.pack_into( givenTrgBuffer, trgOffset, *curValues_LIST )

	########################################
	def GetNameValue_LIST( self ):
		nameValue_TPL_LIST	= range( len( self.mValues_DICT ) )
		for fieldName, structValue in self.mValues_DICT.iteritems():
			nameValue_TPL_LIST[ structValue.mPackTPLidx ] = ( fieldName, structValue.mTheValue )
		return nameValue_TPL_LIST

	########################################
	def GetVisibleNameValue_LIST( self ):
		nameValue_TPL_LIST = range( self.mNumVisible )
		for fieldName, structValue in self.mValues_DICT.iteritems():
			if structValue.mVisibleIdx == None:
				continue
			nameValue_TPL_LIST[ structValue.mVisibleIdx ] = ( fieldName, structValue.mTheValue )
		return nameValue_TPL_LIST

	########################################
	def __repr__( self ):
		nameValue_TPL_LIST = self.GetVisibleNameValue_LIST()
		if len( nameValue_TPL_LIST ) == 0:
			return ' :empty: '
		nameValue_TPL_LIST_itr = iter( nameValue_TPL_LIST )
		nameValue_TPL = nameValue_TPL_LIST_itr.next()
		strRep = "%s: %s" % ( nameValue_TPL[0], nameValue_TPL[1] )
		for nameValue_TPL in nameValue_TPL_LIST_itr:
			strRep = strRep + " %s: %s" % ( nameValue_TPL[0], nameValue_TPL[1] )
		return strRep

####################################################################################################
# END
